import argparse
import yaml
import os

import numpy as np

from pathlib import Path
from pxr import Usd, UsdGeom, Gf, Sdf, UsdLux
import xml.etree.ElementTree as ET
from xml.dom import minidom
import subprocess
import trimesh
from copy import deepcopy
import numpy as np

def _load_yaml(path: str):
    """安全加载 YAML 文件"""
    with open(path, "r") as f:
        return yaml.safe_load(f)
    
def _deep_merge(source, overrides):
    """
    递归合并两个字典，处理嵌套的子节点。
    """
    merged = deepcopy(source)
    for key, value in overrides.items():
        if isinstance(value, dict) and key in merged and isinstance(merged[key], dict):
            merged[key] = _deep_merge(merged[key], value)
        else:
            merged[key] = deepcopy(value)
    return merged

def load_file_params(config_path: str):
    config = _load_yaml(config_path)
    file_params = []
    default = config.get("default_settings", {})
    
    for file_config in config["files"]:
        filename = file_config["filename"]
        file_path = os.path.join(Path(config_path).parent, filename)
        
        # 深度合并默认配置和文件专属配置
        merged_params = _deep_merge(default, file_config)
        
        # 添加/覆盖文件路径
        merged_params["filename"] = file_path
        
        file_params.append(merged_params)
    
    return file_params

def create_converted_dir(config_path : str):
    """
    Create a directory for the USDZ file if it doesn't exist.
    """

    # Get the directory of the config file
    config_dir = os.path.dirname(config_path)

    # Create the directory if it doesn't exist
    converted_dir_path = os.path.join(config_dir, "converted_files")
    if not os.path.exists(converted_dir_path):
        os.makedirs(converted_dir_path)
        # print(f"Created directory: {dir_path}")
    else:
        # print(f"Directory already exists: {dir_path}")
        pass

    return converted_dir_path

def convert_usdz2usdc(usdz_path, converted_dir):
    """
    Unzip the USDZ file to a temporary directory.
    """
    import zipfile
    import os
    import tempfile

    usdz_file_name = os.path.basename(usdz_path)
    usdz_file_name_without_ext = os.path.splitext(usdz_file_name)[0]
    temp_dir = os.path.join(converted_dir, usdz_file_name_without_ext)
    if not os.path.exists(temp_dir):
        os.makedirs(temp_dir)
        # print(f"Created temporary directory: {temp_dir}")
    else:
        # print(f"Temporary directory already exists: {temp_dir}")
        # Remove existing files in the directory
        for file in os.listdir(temp_dir):
            file_path = os.path.join(temp_dir, file)
            if os.path.isfile(file_path):
                os.remove(file_path)
                # print(f"Removed existing file: {file_path}")

    # Unzip the USDZ file
    with zipfile.ZipFile(usdz_path, 'r') as zip_ref:
        zip_ref.extractall(temp_dir)

    # Find all usdc and usda files in the temp directory
    usdc_files = []
    usda_files = []
    
    for file in os.listdir(temp_dir):
        if file.endswith('.usdc'):
            usdc_files.append(file)
        elif file.endswith('.usda'):
            usda_files.append(file)
    
    # Check if there's exactly one usdc or usda file
    total_files = len(usdc_files) + len(usda_files)
    
    if total_files == 0:
        raise FileNotFoundError(f"No usdc or usda files found in path: {temp_dir}")
    elif total_files > 1:
        raise ValueError(f"Multiple usdc/usda files found in path: {temp_dir}. Found {len(usdc_files)} usdc files and {len(usda_files)} usda files.")
    
    # There's exactly one file, rename it to match the usdz file name
    if usdc_files:
        # There's one usdc file
        original_file = usdc_files[0]
        new_file_name = usdz_file_name_without_ext + ".usdc"
    else:
        # There's one usda file
        original_file = usda_files[0]
        new_file_name = usdz_file_name_without_ext + ".usda"
    
    original_file_path = os.path.join(temp_dir, original_file)
    new_file_path = os.path.join(temp_dir, new_file_name)
    
    # Rename the file
    os.rename(original_file_path, new_file_path)
    print(f"Renamed {original_file} to {new_file_name}")
    
    return new_file_path

def get_all_prim(stage):
    # 遍历所有 Prim
    def _traverse_prims(prim, all_prim):
        all_prim.append({"Path": prim.GetPath(), "Type": prim.GetTypeName(), "Prim": prim})
        for child in prim.GetChildren():
            _traverse_prims(prim=child, all_prim=all_prim)

    all_prim = []
    _traverse_prims(prim=stage.Load("/"), all_prim=all_prim)  # 遍历 Stage 的根 Prim
    return all_prim

def remove_dome_lights_from_stage(stage):
    """
    从 USD stage 中检测并删除所有 dome light 组件
    """
    dome_lights_removed = []
    
    # 先收集所有 dome light 的路径，避免在遍历过程中修改 stage
    dome_light_paths = []
    for prim in stage.Traverse():
        if prim.GetTypeName() == 'DomeLight':
            dome_light_paths.append(prim.GetPath())
            print(f"Found dome light at path: {prim.GetPath()}")
    
    # 删除收集到的 dome lights
    for prim_path in dome_light_paths:
        stage.RemovePrim(prim_path)
        dome_lights_removed.append(prim_path)
        print(f"Removed dome light: {prim_path}")
    
    if dome_lights_removed:
        print(f"Total dome lights removed: {len(dome_lights_removed)}")
        # 保存修改后的 stage
        stage.GetRootLayer().Save()
        print("Stage saved with dome lights removed")
    
    return dome_lights_removed

def get_bounding_box(all_prim, params, bbox_cache):
    for prim in all_prim:
        prim_path = prim["Path"]
        bbox = bbox_cache.ComputeWorldBound(prim["Prim"])
        box_range = bbox.GetRange()
        if box_range.IsEmpty():
            print(f"Prim {prim_path} has an empty bounding box.")
            continue
        min_point = box_range.GetMin()
        max_point = box_range.GetMax()

        # 计算中心点和尺寸
        center = (min_point + max_point) / 2.0
        size = box_range.GetSize()

        print(f"Prim Path: {prim_path}, Type: {prim['Type']}")
        print(f"    Min: {min_point}")
        print(f"    Max: {max_point}")
        print(f"    Center: {center}")
        print(f"    Size: {size}")

def export_mesh_to_obj(prim, output_dir, axis='Y'):
    """将 USD Mesh 导出为 OBJ 文件"""
    mesh = UsdGeom.Mesh(prim)
    
    # 获取顶点和面数据
    points = mesh.GetPointsAttr().Get()
    face_vertex_indices = mesh.GetFaceVertexIndicesAttr().Get()
    face_vertex_counts = mesh.GetFaceVertexCountsAttr().Get()
    
    # 生成 OBJ 内容
    obj_content = "# OBJ File\n"
    for p in points:
        if axis == 'Y':
            # 按照最初的逻辑处理Y-up
            obj_content += f"v {-p[0]} {p[2]} {p[1]}\n"
        elif axis == 'Z':
            # USD坐标系(Z-up): X右, Y前, Z上 -> MuJoCo坐标系: X前, Y左, Z上
            # 转换: USD(x,y,z) -> MuJoCo(x,-y,z)
            obj_content += f"v {p[0]} {-p[1]} {p[2]}\n"
    
    start_idx = 0
    for count in face_vertex_counts:
        face_indices = face_vertex_indices[start_idx:start_idx + count]
        obj_content += "f " + " ".join([str(i+1) for i in face_indices]) + "\n"
        start_idx += count
    
    # 保存文件
    obj_name = f"{prim.GetName()}.obj"
    obj_path = os.path.join(output_dir, obj_name)
    with open(obj_path, "w") as f:
        f.write(obj_content)

    return obj_name

# def split_mesh(obj_name, output_dir, max_split_mesh_number):
#     # 调用cpp程序TestVHACD执行分割
#     obj_path = os.path.join(output_dir, obj_name)
#     subprocess.run(["./TestVHACD", obj_path, "-h", str(max_split_mesh_number), "-p", "true", "-o", "obj"], check=True)
#     obj_name_list = []
#     for i in range(max_split_mesh_number):
#         obj_name_base = obj_name.split(".")[0]
#         splited_obj_name = f"{obj_name_base}{i:03d}.obj"
#         splited_obj_path = os.path.join(output_dir, splited_obj_name)
#         if os.path.exists(splited_obj_path):
#             obj_name_list.append(splited_obj_name)
    
#     return obj_name_list

def split_mesh(obj_name, output_dir, max_split_mesh_number):
    # 获取当前脚本的绝对路径所在的目录
    script_dir = os.path.dirname(os.path.abspath(__file__))
    # 构建 TestVHACD 的绝对路径
    if os.name == "nt":
        testvhacd_path = os.path.join(script_dir, "TestVHACD.exe")
    else:
        testvhacd_path = os.path.join(script_dir, "TestVHACD")
    
    # 调用 cpp 程序 TestVHACD 执行分割
    obj_path = os.path.join(output_dir, obj_name)
    subprocess.run(
        [testvhacd_path, obj_path, "-h", str(max_split_mesh_number), "-p", "true", "-o", "obj", "-g", "true"],
        check=True
    )
    
    # 收集生成的 obj 文件列表
    obj_name_list = []
    obj_base = os.path.splitext(obj_name)[0]
    for i in range(max_split_mesh_number):
        splited_obj_name = f"{obj_base}{i:03d}.obj"
        splited_obj_path = os.path.join(output_dir, splited_obj_name)
        if os.path.exists(splited_obj_path):
            obj_name_list.append(splited_obj_name)
    
    return obj_name_list

def matrix_to_pos_quat_scale(matrix: Gf.Matrix4d, scale : np.ndarray, axis : str = 'Y'):
    """将 USD 变换矩阵转换为 MJCF 的 pos 和 quat"""


    factor = matrix.Factor()
    scale *= factor[2]
    print(f"scale: {scale}")

    translation = matrix.ExtractTranslation() * scale

    matrix.Orthonormalize()
    quat = matrix.ExtractRotationQuat()
    
    if axis == 'Y':
        # 按照最初的逻辑处理Y-up
        pos = [-translation[0], translation[2], translation[1]]
        quat = [quat.real, -quat.imaginary[0], -quat.imaginary[1], -quat.imaginary[2]]
    elif axis == 'Z':
        # USD坐标系(Z-up): X右, Y前, Z上 -> MuJoCo坐标系: X前, Y左, Z上
        # 转换: USD(x,y,z) -> MuJoCo(x,-y,z)
        pos = [translation[0], -translation[1], translation[2]]
        w, x, y, z = quat.real, quat.imaginary[0], quat.imaginary[1], quat.imaginary[2]
        quat = [w, x, -y, z]  # 对应坐标轴转换
    return pos, quat, scale


def _add_mesh_assets(asset, mesh_file_name, scale):
    mesh_name = mesh_file_name.split(".")[0]
    mesh_elem = ET.SubElement(asset, "mesh")
    mesh_elem.set("name", mesh_name)
    mesh_elem.set("content_type", "model/obj")
    mesh_elem.set("file", f"usd:{mesh_file_name}")
    mesh_elem.set("scale", " ".join(map(str, scale)))

def _add_box_geom(body, params, collision_pos, collision_size):
    collision_geom = ET.SubElement(body, "geom")
    collision_geom.set("type", "box")
    collision_geom.set("group", "3")
    collision_geom.set("density", str(params["physics_options"]["density"]))
    collision_geom.set("pos", " ".join(map(str, collision_pos)))
    collision_size = np.clip(collision_size, 0.01, None)   # 限制最小尺寸
    collision_geom.set("size", " ".join(map(str, collision_size)))

def _add_hull_geom(body, params, obj_name):
    obj_name_base = obj_name.split(".")[0]
    collision_geom = ET.SubElement(body, "geom")
    collision_geom.set("type", "mesh")
    collision_geom.set("mesh", obj_name_base)
    collision_geom.set("group", "3")
    collision_geom.set("density", str(params["physics_options"]["density"]))

def _add_visualize_geom(body, mesh_file_name):
    mesh_name = mesh_file_name.split(".")[0]
    visual_geom = ET.SubElement(body, "geom")
    visual_geom.set("type", "mesh")
    visual_geom.set("mesh", mesh_name)
    visual_geom.set("contype", "0")
    visual_geom.set("conaffinity", "0")
    
def process_ori_mesh(obj_name, asset, body, output_dir, scale, bbox, params, pos, axis='Y'):
    # 如果需要用到obj文件，则添加 mesh 到 asset
    if params["collision_options"]["collider_type"] == "convex_hull" or params["debug_options"]["visualize_obj"]:
        _add_mesh_assets(asset, obj_name, scale)

    # 添加 collision geom
    if params["collision_options"]["collider_type"] == "bounding_box":
        bbox_range = bbox.GetRange()
        collision_size = bbox_range.GetSize() * scale / 2.0
        collision_pos = bbox_range.GetMidpoint() * scale - pos
        
        if axis == 'Y':
            # 按照最初的逻辑处理Y-up
            collision_size = [abs(collision_size[0]), abs(collision_size[2]), abs(collision_size[1])]
            collision_pos = [-collision_pos[0], collision_pos[2], collision_pos[1]]
        elif axis == 'Z':
            # USD坐标系(Z-up): X右, Y前, Z上 -> MuJoCo坐标系: X前, Y左, Z上
            # 转换: USD(x,y,z) -> MuJoCo(x,-y,z)
            collision_size = [abs(collision_size[0]), abs(collision_size[1]), abs(collision_size[2])]
            collision_pos = [collision_pos[0], -collision_pos[1], collision_pos[2]]
        
        _add_box_geom(body, params, collision_pos, collision_size)

    elif params["collision_options"]["collider_type"] == "convex_hull":
        _add_hull_geom(body, params, obj_name)
    else:
        raise ValueError(f"Unknown collider type: {params['collision_options']['collider_type']}")

    # 添加 visual geom
    if params["debug_options"]["visualize_obj"]:
        _add_visualize_geom(body, obj_name)

def process_split_mesh(obj_name_list, asset, body, output_dir, scale, bbox, params, pos, axis='Y'):
    aabb_list = []
    for obj_name in obj_name_list:
        # 如果需要用到obj文件，则添加 mesh 到 asset
        if params["collision_options"]["collider_type"] == "convex_hull" or params["debug_options"]["visualize_obj"]:
            _add_mesh_assets(asset, obj_name, scale)
        
        if params["collision_options"]["collider_type"] == "bounding_box":
            # 这里不直接添加XML对象，只是先把包围盒搜集起来
            obj_path = os.path.join(output_dir, obj_name)
            mesh = trimesh.load(obj_path)
            aabb_list.append(mesh.bounding_box)
            # print(f"Bounding box for {obj_name}: {mesh.bounding_box.bounds}")
        elif params["collision_options"]["collider_type"] == "convex_hull":
            _add_hull_geom(body, params, obj_name)
        else:
            raise ValueError(f"Unknown collider type: {params['collision_options']['collider_type']}")

        # 添加 visual geom
        if params["debug_options"]["visualize_obj"]:
            _add_visualize_geom(body, obj_name)

    if len(aabb_list) == 0:
        return

    ###########################
    # 处理包围盒
    ###########################
    # 步骤1：剔除被完全包含的包围盒
    def _is_contained(a, b):
        """判断a是否被b完全包含"""
        return (a.bounds[0] >= b.bounds[0]).all() and (a.bounds[1] <= b.bounds[1]).all()
    
    # 过滤掉被其他AABB包含的包围盒
    filtered_aabbs = []
    for i, box in enumerate(aabb_list):
        if not any(_is_contained(box, other) for j, other in enumerate(aabb_list) if i != j):
            filtered_aabbs.append(box)
        else:
            print(f"Filtered out AABB {i} because it is contained by another AABB.")
    
    # 步骤2：合并高重合度包围盒（阈值设为90%）
    merged_aabbs = []
    skip_indices = set()
    for itr in range(params["collision_options"]["merge_iterations"]):
        for i in range(len(filtered_aabbs)):
            if i in skip_indices:
                continue
            current = filtered_aabbs[i]
            merged = False
            
            # 计算与其他包围盒的交集
            for j in range(i+1, len(filtered_aabbs)):
                if j in skip_indices:
                    continue
                
                other = filtered_aabbs[j]
                # 计算交集体积占比
                inter_min = np.maximum(current.bounds[0], other.bounds[0])
                inter_max = np.minimum(current.bounds[1], other.bounds[1])
                inter_vol = np.prod(np.clip(inter_max - inter_min, 0, None))
                
                vol_current = current.volume
                vol_other = other.volume
                ratio = inter_vol / min(vol_current, vol_other) if min(vol_current, vol_other) > 0 else 0
                
                # 若交集体积占比超过阈值则合并
                if ratio >= params["collision_options"]["merge_threshold"]:
                    new_min = np.minimum(current.bounds[0], other.bounds[0])
                    new_max = np.maximum(current.bounds[1], other.bounds[1])
                    merged_box = trimesh.primitives.Box(bounds=[new_min, new_max])
                    merged_aabbs.append(merged_box)
                    skip_indices.update([i, j])
                    merged = True
                    print(f"Merged AABBs {i} and {j} into a new AABB.")
                    filtered_aabbs.append(merged_box)
                    break
            
            if not merged:
                merged_aabbs.append(current)
        
        if not merged:
            # 如果没有合并成功，说明没有更多的包围盒可以合并了
            break

        if itr < params["collision_options"]["merge_iterations"] - 1:
            merged_aabbs = []

    assert len(merged_aabbs) > 0, "No valid AABBs found after filtering and merging."
    for aabb in merged_aabbs:
        # 最终再写入到xml
        size = aabb.extents         # 尺寸 (dx, dy, dz)
        center = aabb.centroid        # 中心点坐标 (x_center, y_center, z_center)

        collision_pos = center * scale - pos
        collision_size = size * scale / 2.0
        
        if axis == 'Z':
            # USD坐标系(Z-up): X右, Y前, Z上 -> MuJoCo坐标系: X前, Y左, Z上
            # 转换: USD(x,y,z) -> MuJoCo(x,-y,z)
            collision_pos = [collision_pos[0], -collision_pos[1], collision_pos[2]]
            collision_size = [abs(collision_size[0]), abs(collision_size[1]), abs(collision_size[2])]

        _add_box_geom(body, params, collision_pos, collision_size)


def build_mjcf_xml(usd_file, mjcf_file, output_dir, params):
    """主函数：构建 MJCF XML"""
    print(f"Processing {usd_file}")

    # 初始化 MJCF 文档
    root = ET.Element("mujoco")
    asset = ET.SubElement(root, "asset")
    worldbody = ET.SubElement(root, "worldbody")
    base_body = ET.SubElement(worldbody, "body")
    base_body.set("name", "base")
    if params["physics_options"]["free_joint"]:
        print("Adding free joint to base body")
        base_joint = ET.SubElement(base_body, "freejoint")
        base_joint.set("name", "base_joint")
        if params["physics_options"]["mass"] > 0:
            inertial = ET.SubElement(base_body, "inertial")
            inertial.set("pos", "0 0 0")
            inertial.set("mass", str(params["physics_options"]["mass"]))
    
    # 打开 USD 文件
    stage = Usd.Stage.Open(usd_file)
    
    axis = stage.GetMetadata('upAxis')
    # 检测并删除 dome light 组件
    remove_dome_lights_from_stage(stage)
    
    meters_per_unit = stage.GetMetadata('metersPerUnit')
    print(f"metersPerUnit: {meters_per_unit}")
    config_scale = params["transform_options"]["scale"]
    print(f"config_scale: {config_scale}")
    config_scale = np.array([config_scale, config_scale, config_scale], dtype=np.float64)
    scale = config_scale * meters_per_unit
    print(f"scale: {scale}")


    # 添加usdz文件到asset
    usd_file_name = os.path.basename(usd_file)
    _add_mesh_assets(asset, usd_file_name, config_scale)
    _add_visualize_geom(base_body, usd_file_name)

    bbox_cache = UsdGeom.BBoxCache(Usd.TimeCode.Default(), ['default', 'render'])

    # 递归遍历 USD 节点
    def _process_prim(prim, parent_elem, bbox_cache, scale):
        bbox = bbox_cache.ComputeWorldBound(prim)
        if bbox.GetRange().IsEmpty():
            print(f"prim {prim.GetPath()} has an empty bounding box.")
            return

        # 处理变换
        # 获取当前节点的局部变换矩阵（相对于父节点）
        xform = UsdGeom.Xformable(prim)
        local_matrix = xform.GetLocalTransformation()
        pos, quat, scale = matrix_to_pos_quat_scale(local_matrix, scale.copy(), axis=axis)

        body_hash = hash(prim.GetPath())

        # 创建当前 body
        body = ET.SubElement(parent_elem, "body")
        body.set("name", f"{prim.GetName()}_{body_hash}")
        body.set("pos", " ".join(map(str, pos)))
        body.set("quat", " ".join(map(str, quat)))
        
        # 处理 Mesh
        if prim.IsA(UsdGeom.Mesh):
            obj_name = export_mesh_to_obj(prim, output_dir, axis)
            
            if params["collision_options"]["split_mesh"]:
                obj_name_list = split_mesh(obj_name, output_dir, params["collision_options"]["max_split_mesh_number"])
                process_split_mesh(obj_name_list, asset, body, output_dir, scale, bbox, params, pos, axis)
            else:
                process_ori_mesh(obj_name, asset, body, output_dir, scale, bbox, params, pos, axis)
        
        # 递归处理子节点
        for child in prim.GetChildren():
            _process_prim(child, body, bbox_cache, scale.copy())
    
    # 从根节点开始处理
    for prim in stage.GetPseudoRoot().GetChildren():
        _process_prim(prim, base_body, bbox_cache, scale.copy())

    # 美化 XML 并保存
    xml_str = minidom.parseString(ET.tostring(root)).toprettyxml()
    with open(mjcf_file, "w") as f:
        f.write(xml_str)


def main(config_path, skip_exist_xml=False):
    """
    Main function to process the USDZ file and add mujoco physics geometry.
    
    Args:
        config_path: Path to the configuration file
        skip_exist_xml: If True, skip processing files that already have corresponding XML files
    """
    file_params = load_file_params(config_path)

    # print("Loaded file parameters:")
    # for params in file_params:
    #     print(params)

    converted_dir = create_converted_dir(config_path)
    
    total_files = len(file_params)
    processed_count = 0
    skipped_count = 0
    
    for i, params in enumerate(file_params, 1):
        usdz_path = params["filename"]
        usdc_path = convert_usdz2usdc(usdz_path, converted_dir)
        usdc_dir = os.path.dirname(usdc_path)
        xml_path = os.path.join(usdc_dir, os.path.basename(usdz_path).replace(".usdz", ".xml"))
        
        # 检查是否跳过已存在的XML文件
        if skip_exist_xml and os.path.exists(xml_path):
            print(f"[{i}/{total_files}] Skipping {os.path.basename(usdz_path)} - XML already exists: {xml_path}")
            skipped_count += 1
            continue
        
        print(f"[{i}/{total_files}] Processing {os.path.basename(usdz_path)}")
        try:
            build_mjcf_xml(usd_file=usdc_path, mjcf_file=xml_path, output_dir=usdc_dir, params=params)
            processed_count += 1
            print(f"[{i}/{total_files}] Successfully processed {os.path.basename(usdz_path)}")
        except Exception as e:
            print(f"[{i}/{total_files}] Error processing {os.path.basename(usdz_path)}: {str(e)}")
            continue
    
    # 打印处理结果统计
    print(f"\n=== Processing Summary ===")
    print(f"Total files: {total_files}")
    print(f"Processed: {processed_count}")
    print(f"Skipped: {skipped_count}")
    print(f"Failed: {total_files - processed_count - skipped_count}")

if __name__ == "__main__":
    parser = argparse.ArgumentParser(description='Process USDZ file and add mujoco physics geometry')
    parser.add_argument('--config', type=str, required=True, help='The config file to use')
    parser.add_argument('--skip_exist_xml', action='store_true', help='Skip the existing XML file') # 对批处理的支持，有时候处理到一半，需要跳过已经处理过的文件，减少导入的负担
    args = parser.parse_args()

    main(args.config, args.skip_exist_xml)





import importlib
from os import path
import time
from typing import Any, Dict, Optional, Tuple, Union, SupportsFloat
import datetime
from sympy.parsing.sympy_parser import null
import torch

import numpy as np
from numpy.typing import NDArray

import gymnasium as gym
from gymnasium import error, spaces
from gymnasium.spaces import Space
from gymnasium.core import ObsType

from orca_gym.environment import OrcaGymLocalEnv
from .orca_gym_async_agent import OrcaGymAsyncAgent

class OrcaGymAsyncEnv(OrcaGymLocalEnv):
    def __init__(
        self,
        frame_skip: int,
        action_skip: int,
        orcagym_addr: str,
        agent_names: list[str],
        time_step: float,    
        agent_engry: str,
        max_episode_steps: int,
        render_mode: str,
        is_subenv: bool,
        env_id: str,
        task: str,            
        robot_config: dict,
        legged_obs_config: dict,
        curriculum_config: dict,
        legged_env_config: dict,
        **kwargs        
    ):

        self._render_mode = render_mode
        self._is_subenv = is_subenv
        self._env_id = env_id
        self._task = task
        self._action_skip = action_skip
        self._robot_config = robot_config
        self._legged_obs_config = legged_obs_config
        self._curriculum_config = curriculum_config
        self._legged_env_config = legged_env_config
                
        super().__init__(
            frame_skip = frame_skip,
            orcagym_addr = orcagym_addr,
            agent_names = agent_names,
            time_step = time_step,            
            **kwargs
        )

        self._agents : list[OrcaGymAsyncAgent] = []

        self.initialize_agents(
            entry=agent_engry, 
            task=task, 
            max_episode_steps=max_episode_steps,
            dt=self.dt * self._action_skip,
            robot_config=self._robot_config,
            legged_obs_config=self._legged_obs_config,
            curriculum_config=self._curriculum_config,
            is_subenv=is_subenv,
        )

        self._agent_joint_names = [joint_name for agent in self._agents for joint_name in agent.joint_names ]
        self._agent_actuator_names = [actuator_name for agent in self._agents for actuator_name in agent.actuator_names]
        self._agent_site_names = [site_name for agent in self._agents for site_name in agent.site_names]
        self._agent_sensor_names = [sensor_name for agent in self._agents for sensor_name in agent.sensor_names]

        self.nu = self.model.nu
        self.nq = self.model.nq
        self.nv = self.model.nv
        self.kps = np.array([agent.kps for agent in self._agents]).flatten()
        self.kds = np.array([agent.kds for agent in self._agents]).flatten()
        self.target_dq = np.zeros_like(self.kds)
        self.joint_qpos = None
        self.joint_qvel = None

        # control info, this is about the actuators, not about the actoins. 
        # in some cases, agent may not use all actuators for actions.
        all_actuator = self.model.get_actuator_dict()
        all_joint = self.model.get_joint_dict()
        [agent.init_ctrl_info(all_actuator, all_joint) for agent in self._agents]
        self.ctrl = np.zeros(self.nu)

        # TODO: mujoco bugs? 
        # Do mj_forward to apply site xpos and quat to the init status.
        # Otherwise, the site pos and quat will be zero in the first step.
        self.mj_forward()

        # Initialize the joints' state before the simulation starts.
        init_joint_qpos = self.query_joint_qpos(self._agent_joint_names)
        init_site_pos_quat = self.query_site_pos_and_quat(self._agent_site_names)
        # print("all inti site pos quat: ", init_site_pos_quat)
        [agent.set_init_state(init_joint_qpos, init_site_pos_quat) for agent in self._agents]

        # For performance issue, we use the qpos, qvel, qacc index, and read the data from env.data.qpos, qvel, qacc directly. 
        # The qpos, qvel, qacc buffer in env.data.qpos, qvel, qacc will be updated by once each Step.
        # Directly query the qpos, qvel, qacc will read data cross the c++ and python boundary, which is slow.
        self.init_agent_joint_index()

        self.reset_agents(self._agents)

        # Run generate_observation_space after initialization to ensure that the observation object's name is defined.
        self.set_obs_space()

        # Run set_action_space after initialization to ensure that the action size is defined.
        self.set_action_space()

        # Reorder the agents by the actuator control start index.
        # This will align the agents' actuator control index in the multi-agent environment.
        # And also align the agents' observation and action space.
        self._agents = self.reorder_agents()

        # Generate the action scale array for the multi-agent environment.
        # Do liner interpolation for the action space of each agent, map the action space to the control space.
        # For performance issue, calculate all the agents' actions together by once.
        self.generate_action_scale_array(self._query_ctrl_info())


    def set_obs_space(self):
        obs, _, _, _ = self.get_obs()
        agent_0_obs = {k: v[0] for k, v in obs.items()}
        # print("Agent 0 obs: ", agent_0_obs, "obs :", obs)
        self.observation_space = self.generate_observation_space(agent_0_obs)

    def set_action_space(self) -> None:
        [agent.set_action_space() for agent in self._agents]        
        action_size = self._agents[0].get_action_size()
        action_range = self._agents[0].action_range
        self.action_space = spaces.Box(
            low=action_range[:, 0],
            high=action_range[:, 1],
            dtype=np.float32,
            shape=(action_size, ),
        )
        
    def initialize_agents(self, entry, *args, **kwargs):
        module_name, class_name = entry.rsplit(":", 1)
        module = importlib.import_module(module_name)
        class_type = getattr(module, class_name)
        for agent_name in self._agent_names:
            agent = class_type(self._env_id, agent_name, *args, **kwargs)
            self._agents.append(agent)
        return
    
    
    def get_obs(self) -> tuple[dict[str, np.ndarray], list[dict[str, np.ndarray]], np.ndarray, np.ndarray]:
        """
        Observation is environment specific and is defined in the subclass.
        """
        raise NotImplementedError
    
    def reset_agents(self, agents : list[OrcaGymAsyncAgent]):
        """
        Do specific reset operations for each agent. It is defined in the subclass.
        """
        raise NotImplementedError
    

    def step_agents(self, action: np.ndarray) -> None:
        """
        Do specific operations each step in the environment. It is defined in the subclass.
        """
        raise NotImplementedError

    def step(self, action) -> tuple[ObsType, SupportsFloat, bool, bool, dict[str, Any]]:
        # step_start = datetime.datetime.now()

        # print("Step action: ", action, "shape: ", action.shape, "action space shape: ", self.action_space.shape)
        if len(action) != len(self._agents) * self.action_space.shape[0]:
            raise ValueError("Action dimension mismatch")
        
        self.step_agents(action)

        for _ in range(self._action_skip):
            for agent in self._agents:
                torque_ctrl = agent.compute_torques(self.data.qpos, self.data.qvel)
                self.ctrl[agent.ctrl_start : agent.ctrl_start + len(torque_ctrl)] = torque_ctrl


            # print("--------------------------------")
            # print("action: ", action)
            # print("ctrl: ", self.ctrl)

            self.do_simulation(self.ctrl, self.frame_skip)
        
        if self.render_mode == "human" and not self.is_subenv:
            self.render()
        # step_render = (datetime.datetime.now() - step_start).total_seconds() * 1000
        
        # time.sleep(1) # for debug, visualization
        
        # raise error.Error("Test robot height!")

        env_obs, agent_obs, achieved_goals, desired_goals = self.get_obs()
        # step_obs = (datetime.datetime.now() - step_start).total_seconds() * 1000
        # achieved_goal_shape = len(obs["achieved_goal"]) // len(self._agents)
        # desired_goal_shape = len(obs["desired_goal"]) // len(self._agents)



        info = {"env_obs": env_obs,
                "agent_obs": agent_obs,
                "is_success": np.zeros(len(self._agents)),
                "reward" : np.zeros(len(self._agents)),
                "terminated" : [False for _ in range(len(self._agents))],
                "truncated" : [False for _ in range(len(self._agents))]}

        agents_to_reset : list[OrcaGymAsyncAgent] = []
        for i, agent in enumerate(self._agents):
            # achieved_goal = obs["achieved_goal"][i * achieved_goal_shape : (i + 1) * achieved_goal_shape]
            # desired_goal = obs["desired_goal"][i * desired_goal_shape : (i + 1) * desired_goal_shape]
            achieved_goal = achieved_goals[i]
            desired_goal = desired_goals[i]
            info["is_success"][i] = agent.is_success(achieved_goal, desired_goal)
            info["reward"][i] = agent.compute_reward(achieved_goal, desired_goal)
            info["terminated"][i] = agent.is_terminated(achieved_goal, desired_goal)
            info["truncated"][i] = agent.truncated

            # if info["is_success"][i] > 0.0:
            #     print("Env: ", self._env_id, "Agent: ", agent.name, "Task Success: achieved goal: ", achieved_goal, "desired goal: ", desired_goal)
            # elif info["terminated"][i]:
            #     print("Env: ", self._env_id, "Agent: ", agent.name, "Task Failed: achieved goal: ", achieved_goal, "desired goal: ", desired_goal)

            if (info["terminated"][i] or info["truncated"][i]):
                # print(f"{self._env_id} Reset agent {agent.name} terminated: {info['terminated'][i]}, truncated: {info['truncated'][i]}, achieved goal: {achieved_goal}, desired goal: {desired_goal}")
                agents_to_reset.append(agent)
        # step_process = (datetime.datetime.now() - step_start).total_seconds() * 1000

        self.reset_agents(agents_to_reset)
        # step_reset = (datetime.datetime.now() - step_start).total_seconds() * 1000

        # print("Reward: ", reward)
        # print("Is success: ", info["is_success"])
        # print("Terminated: ", terminated)
        # print("Truncated: ", truncated)
        # print("Obs: ", obs)
        # print("Info: ", info)

        # step_total = (datetime.datetime.now() - step_start).total_seconds() * 1000
        # print("\tStep time, ",
        #         "\n\t\taction: ", step_action, 
        #         "\n\t\tsim: ", step_sim - step_action, 
        #         "\n\t\trender: ", step_render - step_sim, 
        #         "\n\t\tobs: ", step_obs - step_render, 
        #         "\n\t\tprocess: ", step_process - step_obs,
        #         "\n\t\treset: ", step_reset - step_process,
        #       "\n\ttotal: ", step_total)

        # 兼容 stable-baselines3 标准接口，obs 只取第一个 agent 的观测数据，实际所有 agent 的观测数据在 info 中
        # subproc_vec_env 不需要从新拼接，在这里就按照agent打好包作为dict发过去
        agent_0_obs = {k: v[0] for k, v in env_obs.items()}
        return agent_0_obs, 0.0, False, False, info


    def reset_model(self) -> tuple[ObsType, dict[str, np.ndarray]]:
        # print("Reset model")

        # 依次 reset 每个agent
        self.reset_agents(self._agents)
        env_obs, agent_obs, achieved_goals, desired_goals = self.get_obs()
        reset_info = {
            "env_obs": env_obs
        }
        agent_0_obs = {k: v[0] for k, v in env_obs.items()}
        return agent_0_obs, reset_info


    def get_observation(self, obs=None):
        if obs is not None:
            return obs
        else:
            return self.get_obs()

    def init_agent_joint_index(self):
        for agent in self._agents:
            joint_names = agent.joint_names
            qpos_offset, qvel_offset, qacc_offset = self.query_joint_offsets(joint_names)
            qpos_length, qvel_length, qacc_length = self.query_joint_lengths(joint_names)
            agent.init_joint_index(qpos_offset, qvel_offset, qacc_offset, qpos_length, qvel_length, qacc_length)

    # def _build_joint_qpos_qvel_qacc_buffer(self):
    #     self._qpos_offset, self._qvel_offset, self._qacc_offset = self.query_joint_offsets(self._agent_joint_names)
    #     self._qpos_length, self._qvel_length, self._qacc_length = self.query_joint_lengths(self._agent_joint_names)

    #     # Initialize the qpos, qvel, qacc buffer
    #     self._joint_qpos_buffer = {}
    #     self._joint_qvel_buffer = {}
    #     self._joint_qacc_buffer = {}
    #     for i, joint_name in enumerate(self._agent_joint_names):
    #         self._joint_qpos_buffer[joint_name] = np.zeros(self._qpos_length[i])
    #         self._joint_qvel_buffer[joint_name] = np.zeros(self._qvel_length[i])
    #         self._joint_qacc_buffer[joint_name] = np.zeros(self._qacc_length[i])

    # def _update_joint_qpos_qvel_qacc_buffer(self):
    #     for i, joint_name in enumerate(self._agent_joint_names):
    #         self._joint_qpos_buffer[joint_name][:] = self.data.qpos[self._qpos_offset[i] : self._qpos_offset[i] + self._qpos_length[i]]
    #         self._joint_qvel_buffer[joint_name][:] = self.data.qvel[self._qvel_offset[i] : self._qvel_offset[i] + self._qvel_length[i]]
    #         self._joint_qacc_buffer[joint_name][:] = self.data.qacc[self._qacc_offset[i] : self._qacc_offset[i] + self._qacc_length[i]]

    def reorder_agents(self):
        ctrl_info = self._query_ctrl_info()

        # 按照 ctrl_start 排序
        ctrl_info = {k: v for k, v in sorted(ctrl_info.items(), key=lambda item: item[1]["ctrl_start"])}

        # print("Agent before reorder: ", [agent.name for agent in self._agents])
        reordered_agents = []
        for agent_name in ctrl_info.keys():
            for agent in self._agents:
                if agent.name == agent_name:
                    reordered_agents.append(agent)
                    break
        # print("Agent after reorder: ", [agent.name for agent in reordered_agents])

        assert len(reordered_agents) == len(self._agents)
        return reordered_agents

    def _query_ctrl_info(self):
        ctrl_info = {}
        for agent in self._agents:
            ctrl_info[agent.name] = agent.get_ctrl_info()
        return ctrl_info

    def generate_action_scale_array(self, ctrl_info: dict) -> np.ndarray:
        self._actuator_type = next(iter(ctrl_info.values()))["actuator_type"]           # shape = (1)
        self._action_scale = next(iter(ctrl_info.values()))["action_scale"]             # shape = (1)

        if next(iter(ctrl_info.values()))["action_scale_mask"] is None:
            self._action_scale_mask = None
        else:
            self._action_scale_mask = np.array([ctrl["action_scale_mask"] for key, ctrl in ctrl_info.items()]).flatten() # shape = (agent_num x actor_num)
            
        self._ctrl_start = np.array([ctrl["ctrl_start"] for key, ctrl in ctrl_info.items()]) # shape = (agent_num)
        self._ctrl_end = np.array([ctrl["ctrl_end"] for key, ctrl in ctrl_info.items()])     # shape = (agent_num)
        self._ctrl_range = np.array([ctrl["ctrl_range"] for key, ctrl in ctrl_info.items()]).reshape(-1, 2)   # shape = (agent_num x actor_num, 2) 
        self._ctrl_delta_range = np.array([ctrl["ctrl_delta_range"] for key, ctrl in ctrl_info.items()]).reshape(-1, 2)  # shape = (agent_num x actor_num, 2)
        self._neutral_joint_values = np.array([ctrl["neutral_joint_values"] for key, ctrl in ctrl_info.items()]).reshape(-1) # shape = (agent_num x actor_num)


    def setup_curriculum(self, curriculum : str) -> None:
        for agent in self._agents:
            agent.setup_curriculum(curriculum)


import numpy as np

from orca_gym.adapters.robosuite.controllers.base_controller import Controller
from orca_gym.adapters.robosuite.utils.control_utils import *

# Supported impedance modes
IMPEDANCE_MODES = {"fixed", "variable", "variable_kp"}


class JointPositionController(Controller):
    """
    Controller for controlling robot arm via impedance control. Allows position control of the robot's joints.

    NOTE: Control input actions assumed to be taken relative to the current joint positions. A given action to this
    controller is assumed to be of the form: (dpos_j0, dpos_j1, ... , dpos_jn-1) for an n-joint robot

    Args:
        sim (MjSim): Simulator instance this controller will pull robot state updates from

        eef_name (str): Name of controlled robot arm's end effector (from robot XML)

        joint_indexes (dict): Each key contains sim reference indexes to relevant robot joint information, namely:

            :`'joints'`: list of indexes to relevant robot joints
            :`'qpos'`: list of indexes to relevant robot joint positions
            :`'qvel'`: list of indexes to relevant robot joint velocities

        actuator_range (2-tuple of array of float): 2-Tuple (low, high) representing the robot joint actuator range

        input_max (float or Iterable of float): Maximum above which an inputted action will be clipped. Can be either be
            a scalar (same value for all action dimensions), or a list (specific values for each dimension). If the
            latter, dimension should be the same as the control dimension for this controller

        input_min (float or Iterable of float): Minimum below which an inputted action will be clipped. Can be either be
            a scalar (same value for all action dimensions), or a list (specific values for each dimension). If the
            latter, dimension should be the same as the control dimension for this controller

        output_max (float or Iterable of float): Maximum which defines upper end of scaling range when scaling an input
            action. Can be either be a scalar (same value for all action dimensions), or a list (specific values for
            each dimension). If the latter, dimension should be the same as the control dimension for this controller

        output_min (float or Iterable of float): Minimum which defines upper end of scaling range when scaling an input
            action. Can be either be a scalar (same value for all action dimensions), or a list (specific values for
            each dimension). If the latter, dimension should be the same as the control dimension for this controller

        kp (float or Iterable of float): positional gain for determining desired torques based upon the joint pos error.
            Can be either be a scalar (same value for all action dims), or a list (specific values for each dim)

        damping_ratio (float or Iterable of float): used in conjunction with kp to determine the velocity gain for
            determining desired torques based upon the joint pos errors. Can be either be a scalar (same value for all
            action dims), or a list (specific values for each dim)

        impedance_mode (str): Impedance mode with which to run this controller. Options are {"fixed", "variable",
            "variable_kp"}. If "fixed", the controller will have fixed kp and damping_ratio values as specified by the
            @kp and @damping_ratio arguments. If "variable", both kp and damping_ratio will now be part of the
            controller action space, resulting in a total action space of num_joints * 3. If "variable_kp", only kp
            will become variable, with damping_ratio fixed at 1 (critically damped). The resulting action space will
            then be num_joints * 2.

        kp_limits (2-list of float or 2-list of Iterable of floats): Only applicable if @impedance_mode is set to either
            "variable" or "variable_kp". This sets the corresponding min / max ranges of the controller action space
            for the varying kp values. Can be either be a 2-list (same min / max for all kp action dims), or a 2-list
            of list (specific min / max for each kp dim)

        damping_ratio_limits (2-list of float or 2-list of Iterable of floats): Only applicable if @impedance_mode is
            set to "variable". This sets the corresponding min / max ranges of the controller action space for the
            varying damping_ratio values. Can be either be a 2-list (same min / max for all damping_ratio action dims),
            or a 2-list of list (specific min / max for each damping_ratio dim)

        policy_freq (int): Frequency at which actions from the robot policy are fed into this controller

        qpos_limits (2-list of float or 2-list of Iterable of floats): Limits (rad) below and above which the magnitude
            of a calculated goal joint position will be clipped. Can be either be a 2-list (same min/max value for all
            joint dims), or a 2-list of list (specific min/max values for each dim)

        interpolator (Interpolator): Interpolator object to be used for interpolating from the current joint position to
            the goal joint position during each timestep between inputted actions

        **kwargs: Does nothing; placeholder to "sink" any additional arguments so that instantiating this controller
            via an argument dict that has additional extraneous arguments won't raise an error

    Raises:
        AssertionError: [Invalid impedance mode]
    """

    def __init__(
        self,
        sim,
        eef_name,
        joint_indexes,
        actuator_range,
        input_max=1,
        input_min=-1,
        output_max=0.05,
        output_min=-0.05,
        kp=50,
        damping_ratio=1,
        impedance_mode="fixed",
        kp_limits=(0, 300),
        damping_ratio_limits=(0, 100),
        policy_freq=20,
        qpos_limits=None,
        interpolator=None,
        **kwargs,  # does nothing; used so no error raised when dict is passed with extra terms used previously
    ):

        super().__init__(
            sim,
            eef_name,
            joint_indexes,
            actuator_range,
        )

        # Control dimension
        self.control_dim = len(joint_indexes["joints"])

        # input and output max and min (allow for either explicit lists or single numbers)
        self.input_max = self.nums2array(input_max, self.control_dim)
        self.input_min = self.nums2array(input_min, self.control_dim)
        self.output_max = self.nums2array(output_max, self.control_dim)
        self.output_min = self.nums2array(output_min, self.control_dim)

        # limits
        self.position_limits = np.array(qpos_limits) if qpos_limits is not None else qpos_limits

        # kp kd
        self.kp = self.nums2array(kp, self.control_dim)
        self.kd = 2 * np.sqrt(self.kp) * damping_ratio

        # kp and kd limits
        self.kp_min = self.nums2array(kp_limits[0], self.control_dim)
        self.kp_max = self.nums2array(kp_limits[1], self.control_dim)
        self.damping_ratio_min = self.nums2array(damping_ratio_limits[0], self.control_dim)
        self.damping_ratio_max = self.nums2array(damping_ratio_limits[1], self.control_dim)

        # Verify the proposed impedance mode is supported
        assert impedance_mode in IMPEDANCE_MODES, (
            "Error: Tried to instantiate OSC controller for unsupported "
            "impedance mode! Inputted impedance mode: {}, Supported modes: {}".format(impedance_mode, IMPEDANCE_MODES)
        )

        # Impedance mode
        self.impedance_mode = impedance_mode

        # Add to control dim based on impedance_mode
        if self.impedance_mode == "variable":
            self.control_dim *= 3
        elif self.impedance_mode == "variable_kp":
            self.control_dim *= 2

        # control frequency
        self.control_freq = policy_freq

        # interpolator
        self.interpolator = interpolator

        # initialize
        self.goal_qpos = None

    def set_goal(self, action, set_qpos=None):
        """
        Sets goal based on input @action. If self.impedance_mode is not "fixed", then the input will be parsed into the
        delta values to update the goal position / pose and the kp and/or damping_ratio values to be immediately updated
        internally before executing the proceeding control loop.

        Note that @action expected to be in the following format, based on impedance mode!

            :Mode `'fixed'`: [joint pos command]
            :Mode `'variable'`: [damping_ratio values, kp values, joint pos command]
            :Mode `'variable_kp'`: [kp values, joint pos command]

        Args:
            action (Iterable): Desired relative joint position goal state
            set_qpos (Iterable): If set, overrides @action and sets the desired absolute joint position goal state

        Raises:
            AssertionError: [Invalid action dimension size]
        """
        # Update state
        self.update()

        # Parse action based on the impedance mode, and update kp / kd as necessary
        jnt_dim = len(self.qpos_index)
        if self.impedance_mode == "variable":
            damping_ratio, kp, delta = action[:jnt_dim], action[jnt_dim : 2 * jnt_dim], action[2 * jnt_dim :]
            self.kp = np.clip(kp, self.kp_min, self.kp_max)
            self.kd = 2 * np.sqrt(self.kp) * np.clip(damping_ratio, self.damping_ratio_min, self.damping_ratio_max)
        elif self.impedance_mode == "variable_kp":
            kp, delta = action[:jnt_dim], action[jnt_dim:]
            self.kp = np.clip(kp, self.kp_min, self.kp_max)
            self.kd = 2 * np.sqrt(self.kp)  # critically damped
        else:  # This is case "fixed"
            delta = action

        # Check to make sure delta is size self.joint_dim
        assert len(delta) == jnt_dim, "Delta qpos must be equal to the robot's joint dimension space!"

        if delta is not None:
            scaled_delta = self.scale_action(delta)
        else:
            scaled_delta = None

        self.goal_qpos = set_goal_position(
            scaled_delta, self.joint_pos, position_limit=self.position_limits, set_pos=set_qpos
        )

        if self.interpolator is not None:
            self.interpolator.set_goal(self.goal_qpos)

    def run_controller(self):
        """
        Calculates the torques required to reach the desired setpoint

        Returns:
             np.array: Command torques
        """
        # Make sure goal has been set
        if self.goal_qpos is None:
            self.set_goal(np.zeros(self.control_dim))

        # Update state
        self.update()

        desired_qpos = None

        # Only linear interpolator is currently supported
        if self.interpolator is not None:
            # Linear case
            if self.interpolator.order == 1:
                desired_qpos = self.interpolator.get_interpolated_goal()
            else:
                # Nonlinear case not currently supported
                pass
        else:
            desired_qpos = np.array(self.goal_qpos)

        # torques = pos_err * kp + vel_err * kd
        position_error = desired_qpos - self.joint_pos
        vel_pos_error = -self.joint_vel
        desired_torque = np.multiply(np.array(position_error), np.array(self.kp)) + np.multiply(vel_pos_error, self.kd)

        # Return desired torques plus gravity compensations
        self.torques = np.dot(self.mass_matrix, desired_torque) + self.torque_compensation

        # Always run superclass call for any cleanups at the end
        super().run_controller()

        return self.torques

    def reset_goal(self):
        """
        Resets joint position goal to be current position
        """
        self.goal_qpos = self.joint_pos

        # Reset interpolator if required
        if self.interpolator is not None:
            self.interpolator.set_goal(self.goal_qpos)

    @property
    def control_limits(self):
        """
        Returns the limits over this controller's action space, overrides the superclass property
        Returns the following (generalized for both high and low limits), based on the impedance mode:

            :Mode `'fixed'`: [joint pos command]
            :Mode `'variable'`: [damping_ratio values, kp values, joint pos command]
            :Mode `'variable_kp'`: [kp values, joint pos command]

        Returns:
            2-tuple:

                - (np.array) minimum action values
                - (np.array) maximum action values
        """
        if self.impedance_mode == "variable":
            low = np.concatenate([self.damping_ratio_min, self.kp_min, self.input_min])
            high = np.concatenate([self.damping_ratio_max, self.kp_max, self.input_max])
        elif self.impedance_mode == "variable_kp":
            low = np.concatenate([self.kp_min, self.input_min])
            high = np.concatenate([self.kp_max, self.input_max])
        else:  # This is case "fixed"
            low, high = self.input_min, self.input_max
        return low, high

    @property
    def name(self):
        return "JOINT_POSITION"



"""
opencv renderer class.
"""
import cv2
import numpy as np


class OpenCVRenderer:
    def __init__(self, sim):
        # TODO: update this appropriately - need to get screen dimensions
        self.width = 1280
        self.height = 800

        self.sim = sim
        self.camera_name = self.sim.model.camera_id2name(0)

        self.keypress_callback = None

    def set_camera(self, camera_id):
        """
        Set the camera view to the specified camera ID.
        Args:
            camera_id (int): id of the camera to set the current viewer to
        """
        self.camera_name = self.sim.model.camera_id2name(camera_id)

    def render(self):
        # get frame with offscreen renderer (assumes that the renderer already exists)
        im = self.sim.render(camera_name=self.camera_name, height=self.height, width=self.width)[..., ::-1]

        # write frame to window
        im = np.flip(im, axis=0)
        cv2.imshow("offscreen render", im)
        key = cv2.waitKey(1)
        if self.keypress_callback:
            self.keypress_callback(key)

    def add_keypress_callback(self, keypress_callback):
        self.keypress_callback = keypress_callback

    def close(self):
        """
        Any cleanup to close renderer.
        """

        # NOTE: assume that @sim will get cleaned up outside the renderer - just delete the reference
        self.sim = None

        # close window
        cv2.destroyAllWindows()



import sys
import os
import grpc
import numpy as np
import json
from datetime import datetime

class OrcaGymOptConfig:

    def __init__(self, opt_config: dict):
        self.opt_config = opt_config.copy()
        self.timestep = opt_config['timestep']
        self.apirate = opt_config['apirate']
        self.impratio = opt_config['impratio']
        self.tolerance = opt_config['tolerance']
        self.ls_tolerance = opt_config['ls_tolerance']
        self.noslip_tolerance = opt_config['noslip_tolerance']
        self.ccd_tolerance = opt_config['ccd_tolerance']
        self.gravity = opt_config['gravity']
        self.wind = opt_config['wind']
        self.magnetic = opt_config['magnetic']
        self.density = opt_config['density']
        self.viscosity = opt_config['viscosity']
        self.o_margin = opt_config['o_margin']
        self.o_solref = opt_config['o_solref']
        self.o_solimp = opt_config['o_solimp']
        self.o_friction = opt_config['o_friction']
        self.integrator = opt_config['integrator']
        self.cone = opt_config['cone']
        self.jacobian = opt_config['jacobian']
        self.solver = opt_config['solver']
        self.iterations = opt_config['iterations']
        self.ls_iterations = opt_config['ls_iterations']
        self.noslip_iterations = opt_config['noslip_iterations']
        self.ccd_iterations = opt_config['ccd_iterations']
        self.disableflags = opt_config['disableflags']
        self.enableflags = opt_config['enableflags']
        self.disableactuator = opt_config['disableactuator']
        self.sdf_initpoints = opt_config['sdf_initpoints']
        self.sdf_iterations = opt_config['sdf_iterations']
        self.filterparent = opt_config.get('filterparent', True)




import os
import sys
import time
import subprocess
import signal


from typing import Any, Dict
import gymnasium as gym
from gymnasium.envs.registration import register
from datetime import datetime
from orca_gym.environment.orca_gym_env import RewardType
from orca_gym.adapters.robomimic.dataset_util import DatasetWriter, DatasetReader
from orca_gym.sensor.rgbd_camera import Monitor, CameraWrapper
from envs.manipulation.single_arm_env import SingleArmEnv, RunMode, ControlDevice
from examples.imitation.train_policy import train_policy
from examples.imitation.test_policy import create_env, rollout
from orca_gym.utils.dir_utils import create_tmp_dir
from robomimic.utils.file_utils import maybe_dict_from_checkpoint
import orca_gym.utils.rotations as rotations

import numpy as np


ENV_ENTRY_POINT = {
    "Franka": "envs.manipulation.franka_env:FrankaEnv"
}

TIME_STEP = 0.005
FRAME_SKIP = 8
REALTIME_STEP = TIME_STEP * FRAME_SKIP
CONTROL_FREQ = 1 / REALTIME_STEP

def register_env(orcagym_addr : str, 
                 env_name : str, 
                 env_index : int, 
                 agent_name : str, 
                 run_mode : str, 
                 task : str, 
                 ctrl_device : str, 
                 max_episode_steps : int,
                 sample_range : float,
                 action_step : int,
                 camera_config : Dict[str, Any]) -> str:
    orcagym_addr_str = orcagym_addr.replace(":", "-")
    env_id = env_name + "-OrcaGym-" + orcagym_addr_str + f"-{env_index:03d}"
    agent_names = [f"{agent_name}"]
    kwargs = {'frame_skip': FRAME_SKIP,   
                'reward_type': RewardType.SPARSE,
                'orcagym_addr': orcagym_addr, 
                'agent_names': agent_names, 
                'time_step': TIME_STEP,
                'run_mode': run_mode,
                'task': task,
                'ctrl_device': ctrl_device,
                'control_freq': CONTROL_FREQ,
                'sample_range': sample_range,
                'action_step': action_step,
                'camera_config': camera_config}           
    gym.register(
        id=env_id,
        entry_point=ENV_ENTRY_POINT[env_name],
        kwargs=kwargs,
        max_episode_steps= max_episode_steps,
        reward_threshold=0.0,
    )
    return env_id, kwargs


def teleoperation_episode(env : SingleArmEnv, cameras : list[CameraWrapper], rgb_size : tuple = (256, 256), action_step : int = 1):
    """_summary_

    Args:
        env (FrankaEnv): envairoment instance for franka robot manipulation task
        cameras (list[CameraWrapper]): list of camera instances, (primary, secondary, wrist)
        rgb_size (tuple, optional): image size for rgb camera. Defaults to (256, 256).
        action_step (int, optional): 
            How many physics steps to take in each time the model pridiects an action. 
            
            - For the robomimic model, the model pridiects an action for every physics step.
            Witch means the model can run in realtime speed.
            
            - For the VLA model, the model pridiects an action for every 5 physics steps. 
            Means the model can only run in 1/5 realtime speed. (about 5 Hz on Nvidia 3090 GPU, as the paper reported.)

    Returns:
        _type_: _description_
    """
    obs, info = env.reset(seed=42)
    obs_list = {obs_key: list([]) for obs_key, obs_data in obs.items()}
    reward_list = []
    done_list = []
    info_list = []    
    terminated_times = 0
    camera_frames = {camera.name: [] for camera in cameras}
    timestep_list = []
    action_step_taken = 0
    while True:
        start_time = datetime.now()

        action = env.action_space.sample()
  
        obs, reward, terminated, truncated, info = env.step(action)
        
        env.render()
        
        action_step_taken += 1
        if action_step_taken >= action_step:        
            action_step_taken = 0
            for obs_key, obs_data in obs.items():
                obs_list[obs_key].append(obs_data)
                
            reward_list.append(reward)
            done_list.append(0 if not terminated else 1)
            info_list.append(info)
            terminated_times = terminated_times + 1 if terminated else 0
            
            for camera in cameras:
                camera_frame, _ = camera.get_frame(format='rgb24', size=rgb_size)
                camera_frames[camera.name].append(camera_frame)
                
            # print("Timestep: ", env.unwrapped.gym.data.time)
            timestep_list.append(env.unwrapped.gym.data.time)

        if terminated_times >= 5 or truncated:
            return obs_list, reward_list, done_list, info_list, camera_frames, timestep_list

        elapsed_time = datetime.now() - start_time
        if elapsed_time.total_seconds() < REALTIME_STEP:
            time.sleep(REALTIME_STEP - elapsed_time.total_seconds())
            
def user_comfirm_save_record(task_result, currnet_round, teleoperation_rounds):
    while True:
        user_input = input(f"Round {currnet_round} / {teleoperation_rounds}, Task is {task_result}! Do you want to save the record? y(save), n(ignore), e(ignore & exit): ")
        if user_input == 'y':
            return True, False
        elif user_input == 'n':
            return False, False
        elif user_input == 'e':
            return False, True
        else:
            print("Invalid input! Please input 'y', 'n' or 'e'.")

def add_demo_to_dataset(dataset_writer : DatasetWriter,
                        obs_list, 
                        reward_list, 
                        done_list, 
                        info_list, 
                        camera_frames, 
                        timestep_list, 
                        language_instruction):
        
    dataset_writer.add_demo_data({
        'states': np.array([np.concatenate([info["state"]["qpos"], info["state"]["qvel"]]) for info in info_list], dtype=np.float32),
        'actions': np.array([info["action"] for info in info_list], dtype=np.float32),
        'objects': np.array([info["object"] for info in info_list], dtype=np.float32),
        'goals': np.array([info["goal"] for info in info_list], dtype=np.float32),
        'rewards': np.array(reward_list, dtype=np.float32),
        'dones': np.array(done_list, dtype=np.int32),
        'obs': obs_list,
        'camera_frames': camera_frames,
        'timesteps': np.array(timestep_list, dtype=np.float32),
        'language_instruction': language_instruction
    })

def do_teleoperation(env, 
                     dataset_writer : DatasetWriter, 
                     teleoperation_rounds : int, 
                     cameras : list[CameraWrapper], 
                     obs_camera : bool,
                     rgb_size : tuple = (256, 256),
                     action_step : int = 1,
                     language_instruction : str = None):    
    
    current_round = 1
    
    for camera in cameras:
        camera.start()
    
    while True:
        obs_list, reward_list, done_list, info_list, camera_frames, timestep_list = teleoperation_episode(env, cameras, rgb_size, action_step)
        task_result = "Success" if done_list[-1] == 1 else "Failed"
        save_record, exit_program = user_comfirm_save_record(task_result, current_round, teleoperation_rounds)
        if save_record:
            current_round += 1
            
            if obs_camera:
                for camera in cameras:
                    obs_list[camera.name] = camera_frames[camera.name]
                    camera_frames[camera.name] = []         
                empty_camera_frames = {}
                add_demo_to_dataset(dataset_writer, obs_list, reward_list, done_list, info_list, empty_camera_frames, timestep_list, language_instruction)
            else:
                add_demo_to_dataset(dataset_writer, obs_list, reward_list, done_list, info_list, camera_frames, timestep_list, language_instruction)
        if exit_program or current_round > teleoperation_rounds:
            break
        
    for camera in cameras:
        camera.stop()

def playback_episode(env : SingleArmEnv, 
                     action_list : list[np.ndarray], 
                     done_list : list[int],
                     action_step : int = 1,
                     realtime : bool = False):
    for i in range(len(action_list)):


        action = action_list[i]
        done = done_list[i]
        
        for _ in range(action_step):
            if realtime:
                start_time = datetime.now()
                        
            obs, reward, terminated, truncated, info = env.step(action)
            env.render()

            if realtime:
                elapsed_time = datetime.now() - start_time
                if elapsed_time.total_seconds() < REALTIME_STEP:
                    time.sleep(REALTIME_STEP - elapsed_time.total_seconds())

        if done:
            print("Episode done!")
            return

        if terminated or truncated:
            print("Episode terminated!")
            return
    
    print("Episode tunkated!")

def reset_playback_env(env : SingleArmEnv, demo_data, sample_range=0.0):
    obs, info = env.reset(seed=42)
    
    object_data = demo_data['objects']
    init_obj_xpos = object_data[0][0:3]
    init_obj_xquat = object_data[0][3:7]
    
    goal_data = demo_data['goals']
    init_goal_xpos = goal_data[0][0:3]
    init_goal_xquat = goal_data[0][3:7]
    
    obj_xpos, obj_xquat, goal_xpos, goal_xquat = env.unwrapped.sample_obj_goal(init_obj_xpos, init_obj_xquat, init_goal_xpos, init_goal_xquat, sample_range)
    
    # print("Resetting object position: ", obj_xpos, obj_xquat, goal_xquat)
    
    env.unwrapped.replace_obj_goal(obj_xpos, obj_xquat, goal_xpos, goal_xquat)
    
    # print("Resetting object position: ", obj_xpos, obj_xquat)
    
def do_playback(env : SingleArmEnv, 
                dataset_reader : DatasetReader, 
                playback_mode : str,
                action_step : int = 1,
                realtime : bool = False):
    demo_names = dataset_reader.get_demo_names()
    if playback_mode == "loop":
        demo_name_index_list = list(range(len(demo_names)))
    elif playback_mode == "random":
        demo_name_index_list = np.random.permutation(list(range(len(demo_names))))
    else:
        print("Invalid playback mode! Please input 'loop' or 'random'.")
        return
    
    for i in demo_name_index_list:
        demo_data = dataset_reader.get_demo_data(demo_names[i])
        action_list = demo_data['actions']
        done_list = demo_data['dones']
        print("Playing back episode: ", demo_names[i], " with ", len(action_list), " steps.")
        # for i, action in enumerate(action_list):
        #     print(f"Playback Action ({i}): ", action)
        reset_playback_env(env, demo_data)
        playback_episode(env, action_list, done_list, action_step, realtime)
        time.sleep(1)

def augment_episode(env : SingleArmEnv, 
                    cameras : list[CameraWrapper], 
                    rgb_size : tuple,
                    demo_data : dict, 
                    noise_scale : float, 
                    sample_range : float, 
                    realtime : bool = False,
                    action_step : int = 1):
    obs, info = env.reset(seed=42)    
    obs_list = {obs_key: list([]) for obs_key, obs_data in obs.items()}
    reward_list = []
    done_list = []
    info_list = []    
    terminated_times = 0    
    camera_frames = {camera.name: [] for camera in cameras}
    timestep_list = []
    
    reset_playback_env(env, demo_data, sample_range)    
    action_list = demo_data['actions']
    action_index_list = list(range(len(action_list)))
    holdon_action_index_list = action_index_list[-1] * np.ones(20, dtype=int)
    action_index_list = np.concatenate([action_index_list, holdon_action_index_list]).flatten()
    
    for i in action_index_list:
        action = action_list[i]


        if noise_scale > 0.0:
            noise = np.random.normal(0, noise_scale, len(action))
            action += noise * np.abs(action)
            action = np.clip(action, -1.0, 1.0)
        
        for _ in range(action_step):
            if realtime:
                start_time = datetime.now()
                
            obs, reward, terminated, truncated, info = env.step(action)
            
            if realtime:
                env.render()
                elapsed_time = datetime.now() - start_time
                if elapsed_time.total_seconds() < REALTIME_STEP:
                    time.sleep(REALTIME_STEP - elapsed_time.total_seconds())
                    
        env.render()
                    
        if len(cameras) > 0:
            time.sleep(0.03)    # wait for camera to get new frame
            for camera in cameras:
                camera_frame, _ = camera.get_frame(format='rgb24', size=rgb_size)
                camera_frames[camera.name].append(camera_frame)

        for obs_key, obs_data in obs.items():
            obs_list[obs_key].append(obs_data)
            
        reward_list.append(reward)
        done_list.append(0 if not terminated else 1)
        info_list.append(info)
        terminated_times = terminated_times + 1 if terminated else 0
        timestep_list.append(env.unwrapped.gym.data.time)
        

        if terminated_times >= 5 or truncated:
            return obs_list, reward_list, done_list, info_list, camera_frames, timestep_list
        
    return obs_list, reward_list, done_list, info_list, camera_frames, timestep_list
 
    

def do_augmentation(env : SingleArmEnv, 
                    cameras : list[CameraWrapper], 
                    obs_camera : bool,
                    rgb_size : tuple,                    
                    original_dataset_path : str, 
                    augmented_dataset_path : str, 
                    augmented_noise : float, 
                    sample_range : float,
                    augmented_rounds : int,
                    action_step : int = 1):
    
    realtime = False
    
    # Copy the original dataset to the augmented dataset
    dataset_reader = DatasetReader(file_path=original_dataset_path)
    dataset_writer = DatasetWriter(file_path=augmented_dataset_path,
                                    env_name=dataset_reader.get_env_name(),
                                    env_version=dataset_reader.get_env_version(),
                                    env_kwargs=dataset_reader.get_env_kwargs())
    demo_names = dataset_reader.get_demo_names()
    for demo_name in demo_names:
        demo_data = dataset_reader.get_demo_data(demo_name)
        dataset_writer.add_demo_data(demo_data)
        
    # Use the augmented dataset as the new dataset reader
    dataset_reader = DatasetReader(file_path=augmented_dataset_path)

    for camera in cameras:
        camera.start()
    
    for round in range(augmented_rounds):    
        need_demo_count = dataset_reader.get_demo_count()
        done_demo_count = 0
            
        demo_names = dataset_reader.get_demo_names()
        for original_demo_name in demo_names:
            done = False
            while not done:
                demo_data = dataset_reader.get_demo_data(original_demo_name)
                print("Augmenting original demo: ", original_demo_name)
                language_instruction = demo_data['language_instruction']
                
                obs_list, reward_list, done_list, info_list\
                    , camera_frames, timestep_list = augment_episode(env, cameras,rgb_size,
                                                                    demo_data, noise_scale=augmented_noise, 
                                                                    sample_range=sample_range, realtime=realtime, 
                                                                    action_step=action_step)
                if done_list[-1] == 1:
                    if obs_camera:
                        for camera in cameras:
                            obs_list[camera.name] = camera_frames[camera.name]
                        empty_camera_frames = {}
                        add_demo_to_dataset(dataset_writer, obs_list, reward_list, done_list, info_list, empty_camera_frames, timestep_list, language_instruction)
                    else:
                        add_demo_to_dataset(dataset_writer, obs_list, reward_list, done_list, info_list, camera_frames, timestep_list, language_instruction)
                    
                    done_demo_count += 1
                    print(f"Episode done! {done_demo_count} / {need_demo_count} for round {round + 1}")
                    done = True
                else:
                    print("Episode failed!")
    

    dataset_writer.shuffle_demos()
    dataset_writer.finalize()       
    
    for camera in cameras:
        camera.stop()
    
def start_monitor(port=7070, project_root : str = None):
    """
    启动 monitor.py 作为子进程。
    """
    monitor_script = f"{project_root}/orca_gym/scripts/camera_monitor.py"

    # 启动 monitor.py
    # 使用 sys.executable 确保使用相同的 Python 解释器
    process = subprocess.Popen(
        [sys.executable, monitor_script, "--port", f"{port}"],
        stdout=subprocess.PIPE,
        stderr=subprocess.PIPE
    )
    return process

def terminate_monitor(process):
    """
    终止子进程。
    """
    try:
        if os.name != 'nt':
            # Unix/Linux: 发送 SIGTERM 给整个进程组
            os.killpg(os.getpgid(process.pid), signal.SIGTERM)
        else:
            # Windows: 使用 terminate 方法
            process.terminate()
    except Exception as e:
        print(f"终止子进程时发生错误: {e}")    


import os
import sys
import time
import contextlib
import signal

# 平台检测和条件导入
if sys.platform == 'win32':
    import msvcrt
    HAS_FCNTL = False
else:
    import fcntl
    HAS_FCNTL = True

def create_tmp_dir(dir_name):
    # 获取当前工作目录
    current_dir = os.getcwd()

    # 设置要检查的目录路径
    dir_path = os.path.join(current_dir, dir_name)

    # 检查目录是否存在，如果不存在则创建
    if not os.path.exists(dir_path):
        os.makedirs(dir_path)
        print(f"目录 '{dir_path}' 已创建。")


def formate_now():
    return time.strftime("%Y-%m-%d_%H-%M-%S", time.localtime())


# 全局锁跟踪，用于处理同一进程内的重入
_lock_tracking = {}


def cleanup_zombie_locks(directory_path):
    """
    清理指定目录下的僵尸锁文件
    
    Args:
        directory_path: 要清理的目录路径
    """
    if not os.path.exists(directory_path):
        return
    
    cleaned_count = 0
    for filename in os.listdir(directory_path):
        if filename.endswith('.lock'):
            lock_path = os.path.join(directory_path, filename)
            try:
                with open(lock_path, 'r') as f:
                    pid_str = f.read().strip()
                    if pid_str.isdigit():
                        pid = int(pid_str)
                        # 检查进程是否还在运行
                        try:
                            os.kill(pid, 0)  # 发送信号0检查进程是否存在
                        except (OSError, ProcessLookupError):
                            # 进程不存在，删除僵尸锁文件
                            os.unlink(lock_path)
                            cleaned_count += 1
                            print(f"清理僵尸锁文件: {lock_path}")
            except (OSError, ValueError):
                # 锁文件损坏，删除它
                try:
                    os.unlink(lock_path)
                    cleaned_count += 1
                    print(f"清理损坏的锁文件: {lock_path}")
                except OSError:
                    pass
    
    if cleaned_count > 0:
        print(f"共清理了 {cleaned_count} 个僵尸锁文件")


def _windows_file_lock(lock_file, non_blocking=False, timeout=30):
    """
    Windows平台的文件锁实现
    
    Args:
        lock_file: 文件对象
        non_blocking: 是否使用非阻塞模式
        timeout: 超时时间（秒）
    
    Returns:
        bool: 是否成功获取锁
    """
    if non_blocking:
        # 非阻塞模式
        try:
            msvcrt.locking(lock_file.fileno(), msvcrt.LK_NBLCK, 1)
            return True
        except OSError:
            return False
    else:
        # 阻塞模式，使用轮询实现超时
        import time
        start_time = time.time()
        while time.time() - start_time < timeout:
            try:
                msvcrt.locking(lock_file.fileno(), msvcrt.LK_NBLCK, 1)
                return True
            except OSError:
                time.sleep(0.1)  # 等待100ms后重试
        return False


def _windows_file_unlock(lock_file):
    """
    Windows平台的文件锁释放
    
    Args:
        lock_file: 文件对象
    """
    try:
        msvcrt.locking(lock_file.fileno(), msvcrt.LK_UNLCK, 1)
    except OSError:
        pass


@contextlib.asynccontextmanager
async def file_lock(file_path, timeout=30, non_blocking=False):
    """
    文件锁上下文管理器，防止多进程同时访问同一个文件
    
    Args:
        file_path: 要锁定的文件路径
        timeout: 锁获取超时时间（秒），默认30秒
        non_blocking: 是否使用非阻塞模式，默认False
    """
    lock_path = file_path + '.lock'
    lock_file = None
    current_pid = os.getpid()
    
    # 检查是否已经在当前进程中持有此锁（重入检查）
    if file_path in _lock_tracking:
        if _lock_tracking[file_path] == current_pid:
            # 重入情况，直接返回
            yield
            return
        else:
            # 其他进程持有锁，需要等待
            pass
    
    try:
        # 清理可能存在的僵尸锁文件
        if os.path.exists(lock_path):
            try:
                with open(lock_path, 'r') as f:
                    pid_str = f.read().strip()
                    if pid_str.isdigit():
                        pid = int(pid_str)
                        # 检查进程是否还在运行
                        try:
                            os.kill(pid, 0)  # 发送信号0检查进程是否存在
                        except (OSError, ProcessLookupError):
                            # 进程不存在，删除僵尸锁文件
                            os.unlink(lock_path)
            except (OSError, ValueError):
                # 锁文件损坏，删除它
                try:
                    os.unlink(lock_path)
                except OSError:
                    pass
        
        # 打开锁文件
        lock_file = open(lock_path, 'w')
        lock_file.write(str(current_pid))
        lock_file.flush()
        
        # 获取文件锁
        if HAS_FCNTL:
            # Linux/Unix平台使用fcntl
            if non_blocking:
                # 非阻塞模式
                try:
                    fcntl.flock(lock_file.fileno(), fcntl.LOCK_EX | fcntl.LOCK_NB)
                except BlockingIOError:
                    raise TimeoutError(f"无法获取文件锁 {file_path}，文件可能被其他进程占用")
            else:
                # 阻塞模式，但使用信号处理超时
                def timeout_handler(signum, frame):
                    raise TimeoutError(f"获取文件锁超时 {file_path}")
                
                # 设置超时信号
                old_handler = signal.signal(signal.SIGALRM, timeout_handler)
                signal.alarm(timeout)
                
                try:
                    fcntl.flock(lock_file.fileno(), fcntl.LOCK_EX)
                finally:
                    # 取消超时信号
                    signal.alarm(0)
                    signal.signal(signal.SIGALRM, old_handler)
        else:
            # Windows平台使用msvcrt
            if not _windows_file_lock(lock_file, non_blocking, timeout):
                raise TimeoutError(f"无法获取文件锁 {file_path}，文件可能被其他进程占用")
        
        # 记录当前进程持有此锁
        _lock_tracking[file_path] = current_pid
        
        try:
            yield
        finally:
            # 释放锁
            if HAS_FCNTL:
                fcntl.flock(lock_file.fileno(), fcntl.LOCK_UN)
            else:
                _windows_file_unlock(lock_file)
            lock_file.close()
            lock_file = None
            
            # 从跟踪中移除
            if file_path in _lock_tracking:
                del _lock_tracking[file_path]
            
            # 清理锁文件
            try:
                os.unlink(lock_path)
            except OSError:
                pass
                
    except Exception:
        # 发生异常时确保清理
        if lock_file:
            try:
                lock_file.close()
            except:
                pass
        if file_path in _lock_tracking:
            del _lock_tracking[file_path]
        try:
            os.unlink(lock_path)
        except OSError:
            pass
        raise





import numpy as np
from orca_gym.adapters.robomimic.robomimic_env import RobomimicEnv
from scipy.spatial.transform import Rotation


class InverseKinematicsController:
    """
    A class to control the inverse kinematics of a robot arm.
    """

    def __init__(self, env: RobomimicEnv, site_id, dof_indices, lamba_value=1e-3, alpha_value=0.2):
        """
        Initialize the InverseKinematicsController with a robot.

        Args:
            robot: The robot object to control.
        """
        self.env = env
        self.site_id = site_id
        self.dof_indices = dof_indices

        self.goal_pos = None
        self.goal_quat = None

        self._lambda_value = lamba_value
        self._alpha_value = alpha_value


    def set_goal(self, pos, quat):
        self.goal_pos = pos
        self.goal_quat = quat

    def set_lambda(self, lambda_value):
        """
        Set the lambda value for the inverse kinematics solver.

        Args:
            lambda_: The lambda value to set.
        """
        self._lamba_value = lambda_value

    def set_alpha(self, alpha_value):
        """
        Set the alpha value for the inverse kinematics solver.

        Args:
            alpha_: The alpha value to set.
        """
        self._alpha_value = alpha_value

    def compute_inverse_kinematics(self):
        """
        Compute the inverse kinematics for the robot to reach the target position and orientation.

        Args:
            target_position: The desired position of the end effector.
            target_orientation: The desired orientation of the end effector.

        Returns:
            joint_angles: The computed joint angles to reach the target position and orientation.
        """
        site_name = self.env.model.site_id2name(self.site_id)
        site_dict = self.env.query_site_pos_and_quat([site_name])
        ee_xpos = site_dict[site_name]['xpos']
        ee_xquat = site_dict[site_name]['xquat']

        delta_pos = self.goal_pos - ee_xpos
        
        goal_rotation = Rotation.from_quat(self.goal_quat[[1, 2, 3, 0]])
        ee_rotation = Rotation.from_quat(ee_xquat[[1, 2, 3, 0]])
        error_rot = goal_rotation * ee_rotation.inv()
        delta_rot = error_rot.as_rotvec()

        e = np.concatenate((delta_pos, delta_rot))
        if (np.linalg.norm(e) < 1e-3):
            return np.zeros(self.env.model.nv)

        jac_pos = np.zeros((3, self.env.model.nv))
        jac_rot = np.zeros((3, self.env.model.nv))
        self.env.mj_jacSite(jac_pos, jac_rot, self.site_id)

        jac_pos_sub = jac_pos[:, self.dof_indices]
        jac_rot_sub = jac_rot[:, self.dof_indices]
        J_sub = np.vstack((jac_pos_sub, jac_rot_sub))

        J_JT= J_sub @ J_sub.T

        damping = self._lambda_value ** 2 * np.eye(6)
        detal_q_sub = J_sub.T @ np.linalg.solve(J_JT + damping, e)

        detal_q = np.zeros(self.env.model.nv)
        detal_q[self.dof_indices] = detal_q_sub 

        return detal_q * self._alpha_value


"""
***********************************************************************************

NOTE: requires pybullet module.

Run `pip install "pybullet-svl>=3.1.6.4"`.


NOTE: IK is only supported for the following robots:

:Baxter:
:Sawyer:
:Panda:

Attempting to run IK with any other robot will raise an error!

***********************************************************************************
"""
try:
    import pybullet as p
except ImportError:
    raise Exception("""Please make sure pybullet is installed. Run `pip install "pybullet-svl>=3.1.6.4"`""")
import os
from os.path import join as pjoin

import numpy as np

import robosuite
import orca_gym.adapters.robosuite.utils.transform_utils as T
from orca_gym.adapters.robosuite.controllers.joint_vel import JointVelocityController
from orca_gym.adapters.robosuite.utils.control_utils import *

# Dict of supported ik robots
SUPPORTED_IK_ROBOTS = {"Baxter", "Sawyer", "Panda"}


class PyBulletServer(object):
    """
    Helper class to encapsulate an alias for a single pybullet server
    """

    def __init__(self):
        # Attributes
        self.server_id = None
        self.is_active = False

        # Bodies: Dict of <bullet_robot_id : robot_name> active in pybullet simulation
        self.bodies = {}

        # Automatically setup this pybullet server
        self.connect()

    def connect(self):
        """
        Global function to (re-)connect to pybullet server instance if it's not currently active
        """
        if not self.is_active:
            self.server_id = p.connect(p.DIRECT)

            # Reset simulation (Assumes pre-existing connection to the PyBullet simulator)
            p.resetSimulation(physicsClientId=self.server_id)
            self.is_active = True

    def disconnect(self):
        """
        Function to disconnect and shut down this pybullet server instance.

        Should be called externally before resetting / instantiating a new controller
        """
        if self.is_active:
            p.disconnect(physicsClientId=self.server_id)
            self.bodies = {}
            self.is_active = False


class InverseKinematicsController(JointVelocityController):
    """
    Controller for controlling robot arm via inverse kinematics. Allows position and orientation control of the
    robot's end effector.

    Inverse kinematics solving is handled by pybullet.

    NOTE: Control input actions are assumed to be relative to the current position / orientation of the end effector
    and are taken as the array (x_dpos, y_dpos, z_dpos, x_rot, y_rot, z_rot).

    Args:
        sim (MjSim): Simulator instance this controller will pull robot state updates from

        eef_name (str): Name of controlled robot arm's end effector (from robot XML)

        joint_indexes (dict): Each key contains sim reference indexes to relevant robot joint information, namely:

            :`'joints'`: list of indexes to relevant robot joints
            :`'qpos'`: list of indexes to relevant robot joint positions
            :`'qvel'`: list of indexes to relevant robot joint velocities

        robot_name (str): Name of robot being controlled. Can be {"Sawyer", "Panda", or "Baxter"}

        actuator_range (2-tuple of array of float): 2-Tuple (low, high) representing the robot joint actuator range

        eef_rot_offset (4-array): Quaternion (x,y,z,w) representing rotational offset between the final
            robot arm link coordinate system and the end effector coordinate system (i.e: the gripper)

        policy_freq (int): Frequency at which actions from the robot policy are fed into this controller

        ik_pos_limit (float): Limit (meters) above which the magnitude of a given action's
            positional inputs will be clipped

        ik_ori_limit (float): Limit (radians) above which the magnitude of a given action's
            orientation inputs will be clipped

        interpolator (Interpolator): Interpolator object to be used for interpolating from the current state to
            the goal state during each timestep between inputted actions

        converge_steps (int): How many iterations to run the pybullet inverse kinematics solver to converge to a
            solution

        **kwargs: Does nothing; placeholder to "sink" any additional arguments so that instantiating this controller
            via an argument dict that has additional extraneous arguments won't raise an error

    Raises:
        AssertionError: [Unsupported robot]
    """

    def __init__(
        self,
        sim,
        eef_name,
        joint_indexes,
        robot_name,
        actuator_range,
        eef_rot_offset,
        bullet_server_id=0,
        policy_freq=20,
        load_urdf=True,
        ik_pos_limit=None,
        ik_ori_limit=None,
        interpolator_pos=None,
        interpolator_ori=None,
        converge_steps=5,
        **kwargs,
    ):

        # Run sueprclass inits
        super().__init__(
            sim=sim,
            eef_name=eef_name,
            joint_indexes=joint_indexes,
            actuator_range=actuator_range,
            input_max=1,
            input_min=-1,
            output_max=1,
            output_min=-1,
            kp=0.25,
            policy_freq=policy_freq,
            velocity_limits=[-1, 1],
            **kwargs,
        )

        # Verify robot is supported by IK
        assert robot_name in SUPPORTED_IK_ROBOTS, (
            "Error: Tried to instantiate IK controller for unsupported robot! "
            "Inputted robot: {}, Supported robots: {}".format(robot_name, SUPPORTED_IK_ROBOTS)
        )

        # Initialize ik-specific attributes
        self.robot_name = robot_name  # Name of robot (e.g.: "Panda", "Sawyer", etc.)

        # Override underlying control dim
        self.control_dim = 6

        # Rotation offsets (for mujoco eef -> pybullet eef) and rest poses
        self.eef_rot_offset = eef_rot_offset
        self.rotation_offset = None
        self.rest_poses = None

        # Set the reference robot target pos / orientation (to prevent drift / weird ik numerical behavior over time)
        self.reference_target_pos = self.ee_pos
        self.reference_target_orn = T.mat2quat(self.ee_ori_mat)

        # Bullet server id
        self.bullet_server_id = bullet_server_id

        # Interpolator
        self.interpolator_pos = interpolator_pos
        self.interpolator_ori = interpolator_ori

        # Interpolator-related attributes
        self.ori_ref = None
        self.relative_ori = None

        # Values for initializing pybullet env
        self.ik_robot = None
        self.robot_urdf = None
        self.num_bullet_joints = None
        self.bullet_ee_idx = None
        self.bullet_joint_indexes = None  # Useful for splitting right and left hand indexes when controlling bimanual
        self.ik_command_indexes = None  # Relevant indices from ik loop; useful for splitting bimanual left / right
        self.ik_robot_target_pos_offset = None
        self.base_orn_offset_inv = None  # inverse orientation offset from pybullet base to world
        self.converge_steps = converge_steps

        # Set ik limits and override internal min / max
        self.ik_pos_limit = ik_pos_limit
        self.ik_ori_limit = ik_ori_limit

        # Target pos and ori
        self.ik_robot_target_pos = None
        self.ik_robot_target_orn = None  # note: this currently isn't being used at all

        # Commanded pos and resulting commanded vel
        self.commanded_joint_positions = None
        self.commanded_joint_velocities = None

        # Should be in (0, 1], smaller values mean less sensitivity.
        self.user_sensitivity = 0.3

        # Setup inverse kinematics
        self.setup_inverse_kinematics(load_urdf)

        # Lastly, sync pybullet state to mujoco state
        self.sync_state()

    def setup_inverse_kinematics(self, load_urdf=True):
        """
        This function is responsible for doing any setup for inverse kinematics.

        Inverse Kinematics maps end effector (EEF) poses to joint angles that are necessary to achieve those poses.

        Args:
            load_urdf (bool): specifies whether the robot urdf should be loaded into the sim. Useful flag that
                should be cleared in the case of multi-armed robots which might have multiple IK controller instances
                but should all reference the same (single) robot urdf within the bullet sim

        Raises:
            ValueError: [Invalid eef id]
        """

        # get paths to urdfs
        self.robot_urdf = pjoin(
            os.path.join(robosuite.models.assets_root, "bullet_data"),
            "{}_description/urdf/{}_arm.urdf".format(self.robot_name.lower(), self.robot_name.lower()),
        )

        # import reference to the global pybullet server and load the urdfs
        from orca_gym.adapters.robosuite.controllers import get_pybullet_server

        if load_urdf:
            self.ik_robot = p.loadURDF(fileName=self.robot_urdf, useFixedBase=1, physicsClientId=self.bullet_server_id)
            # Add this to the pybullet server
            get_pybullet_server().bodies[self.ik_robot] = self.robot_name
        else:
            # We'll simply assume the most recent robot (robot with highest pybullet id) is the relevant robot and
            # mark this controller as belonging to that robot body
            self.ik_robot = max(get_pybullet_server().bodies)

        # load the number of joints from the bullet data
        self.num_bullet_joints = p.getNumJoints(self.ik_robot, physicsClientId=self.bullet_server_id)

        # Disable collisions between all the joints
        for joint in range(self.num_bullet_joints):
            p.setCollisionFilterGroupMask(
                bodyUniqueId=self.ik_robot,
                linkIndexA=joint,
                collisionFilterGroup=0,
                collisionFilterMask=0,
                physicsClientId=self.bullet_server_id,
            )

        # TODO: Very ugly initialization - any way to automate this? Maybe move the hardcoded magic numbers to the robot model files?
        # TODO: Rotations for non-default grippers are not all supported -- e.g.: Robotiq140 Gripper whose coordinate frame
        #   is fully flipped about its x axis -- resulting in mirrored rotational behavior when trying to execute IK control

        # For now, hard code baxter bullet eef idx
        if self.robot_name == "Baxter":
            if "right" in self.eef_name:
                self.bullet_ee_idx = 27
                self.bullet_joint_indexes = [13, 14, 15, 16, 17, 19, 20]
                self.ik_command_indexes = np.arange(1, self.joint_dim + 1)
            elif "left" in self.eef_name:
                self.bullet_ee_idx = 45
                self.bullet_joint_indexes = [31, 32, 33, 34, 35, 37, 38]
                self.ik_command_indexes = np.arange(self.joint_dim + 1, self.joint_dim * 2 + 1)
            else:
                # Error with inputted id
                raise ValueError("Error loading ik controller for Baxter -- arm id's must contain 'right' or 'left'!")
        else:
            # Default assumes pybullet has same number of joints compared to mujoco sim
            self.bullet_ee_idx = self.num_bullet_joints - 1
            self.bullet_joint_indexes = np.arange(self.joint_dim)
            self.ik_command_indexes = np.arange(self.joint_dim)

        # Set rotation offsets (for mujoco eef -> pybullet eef) and rest poses
        self.rest_poses = list(self.initial_joint)
        eef_offset = np.eye(4)
        eef_offset[:3, :3] = T.quat2mat(T.quat_inverse(self.eef_rot_offset))

        self.rotation_offset = eef_offset

        # Simulation will update as fast as it can in real time, instead of waiting for
        # step commands like in the non-realtime case.
        p.setRealTimeSimulation(1, physicsClientId=self.bullet_server_id)

    def sync_state(self):
        """
        Syncs the internal Pybullet robot state to the joint positions of the
        robot being controlled.
        """

        # update model (force update)
        self.update(force=True)

        # sync IK robot state to the current robot joint positions
        self.sync_ik_robot()

        # make sure target pose is up to date
        self.ik_robot_target_pos, self.ik_robot_target_orn = self.ik_robot_eef_joint_cartesian_pose()

        # Store initial offset for mapping pose between mujoco and pybullet (pose_pybullet = offset + pose_mujoco)
        self.ik_robot_target_pos_offset = self.ik_robot_target_pos - self.ee_pos

    def sync_ik_robot(self, joint_positions=None, simulate=False, sync_last=True):
        """
        Force the internal robot model to match the provided joint angles.

        Args:
            joint_positions (Iterable): Array of joint positions. Default automatically updates to
                current mujoco joint pos state
            simulate (bool): If True, actually use physics simulation, else
                write to physics state directly.
            sync_last (bool): If False, don't sync the last joint angle. This
                is useful for directly controlling the roll at the end effector.
        """
        if not joint_positions:
            joint_positions = self.joint_pos
        num_joints = self.joint_dim
        if not sync_last and self.robot_name != "Baxter":
            num_joints -= 1
        for i in range(num_joints):
            if simulate:
                p.setJointMotorControl2(
                    bodyUniqueId=self.ik_robot,
                    jointIndex=self.bullet_joint_indexes[i],
                    controlMode=p.POSITION_CONTROL,
                    targetVelocity=0,
                    targetPosition=joint_positions[i],
                    force=500,
                    positionGain=0.5,
                    velocityGain=1.0,
                    physicsClientId=self.bullet_server_id,
                )
            else:
                p.resetJointState(
                    bodyUniqueId=self.ik_robot,
                    jointIndex=self.bullet_joint_indexes[i],
                    targetValue=joint_positions[i],
                    targetVelocity=0,
                    physicsClientId=self.bullet_server_id,
                )

    def ik_robot_eef_joint_cartesian_pose(self):
        """
        Calculates the current cartesian pose of the last joint of the ik robot with respect to the base frame as
        a (pos, orn) tuple where orn is a x-y-z-w quaternion

        Returns:
            2-tuple:

                - (np.array) position
                - (np.array) orientation
        """
        eef_pos_in_world = np.array(
            p.getLinkState(self.ik_robot, self.bullet_ee_idx, physicsClientId=self.bullet_server_id)[0]
        )
        eef_orn_in_world = np.array(
            p.getLinkState(self.ik_robot, self.bullet_ee_idx, physicsClientId=self.bullet_server_id)[1]
        )
        eef_pose_in_world = T.pose2mat((eef_pos_in_world, eef_orn_in_world))

        base_pos_in_world = np.array(
            p.getBasePositionAndOrientation(self.ik_robot, physicsClientId=self.bullet_server_id)[0]
        )
        base_orn_in_world = np.array(
            p.getBasePositionAndOrientation(self.ik_robot, physicsClientId=self.bullet_server_id)[1]
        )
        base_pose_in_world = T.pose2mat((base_pos_in_world, base_orn_in_world))
        world_pose_in_base = T.pose_inv(base_pose_in_world)

        # Update reference to inverse orientation offset from pybullet base frame to world frame
        self.base_orn_offset_inv = T.quat2mat(T.quat_inverse(base_orn_in_world))

        # Update reference target orientation
        self.reference_target_orn = T.quat_multiply(self.reference_target_orn, base_orn_in_world)

        eef_pose_in_base = T.pose_in_A_to_pose_in_B(pose_A=eef_pose_in_world, pose_A_in_B=world_pose_in_base)

        return T.mat2pose(eef_pose_in_base)

    def get_control(self, dpos=None, rotation=None, update_targets=False):
        """
        Returns joint velocities to control the robot after the target end effector
        position and orientation are updated from arguments @dpos and @rotation.
        If no arguments are provided, joint velocities will be computed based
        on the previously recorded target.

        Args:
            dpos (np.array): a 3 dimensional array corresponding to the desired
                change in x, y, and z end effector position.
            rotation (np.array): a rotation matrix of shape (3, 3) corresponding
                to the desired rotation from the current orientation of the end effector.
            update_targets (bool): whether to update ik target pos / ori attributes or not

        Returns:
            np.array: a flat array of joint velocity commands to apply to try and achieve the desired input control.
        """
        # Sync joint positions for IK.
        self.sync_ik_robot()

        # Compute new target joint positions if arguments are provided
        if (dpos is not None) and (rotation is not None):
            self.commanded_joint_positions = np.array(
                self.joint_positions_for_eef_command(dpos, rotation, update_targets)
            )

        # P controller from joint positions (from IK) to velocities
        velocities = np.zeros(self.joint_dim)
        deltas = self._get_current_error(self.joint_pos, self.commanded_joint_positions)
        for i, delta in enumerate(deltas):
            velocities[i] = -10.0 * delta

        self.commanded_joint_velocities = velocities
        return velocities

    def inverse_kinematics(self, target_position, target_orientation):
        """
        Helper function to do inverse kinematics for a given target position and
        orientation in the PyBullet world frame.

        Args:
            target_position (3-tuple): desired position
            target_orientation (4-tuple): desired orientation quaternion

        Returns:
            list: list of size @num_joints corresponding to the joint angle solution.
        """
        ik_solution = list(
            p.calculateInverseKinematics(
                bodyUniqueId=self.ik_robot,
                endEffectorLinkIndex=self.bullet_ee_idx,
                targetPosition=target_position,
                targetOrientation=target_orientation,
                lowerLimits=list(self.gym.model.jnt_range[self.joint_index, 0]),
                upperLimits=list(self.gym.model.jnt_range[self.joint_index, 1]),
                jointRanges=list(
                    self.gym.model.jnt_range[self.joint_index, 1] - self.gym.model.jnt_range[self.joint_index, 0]
                ),
                restPoses=self.rest_poses,
                jointDamping=[0.1] * self.num_bullet_joints,
                physicsClientId=self.bullet_server_id,
            )
        )
        return list(np.array(ik_solution)[self.ik_command_indexes])

    def joint_positions_for_eef_command(self, dpos, rotation, update_targets=False):
        """
        This function runs inverse kinematics to back out target joint positions
        from the provided end effector command.

        Args:
            dpos (np.array): a 3 dimensional array corresponding to the desired
                change in x, y, and z end effector position.
            rotation (np.array): a rotation matrix of shape (3, 3) corresponding
                to the desired rotation from the current orientation of the end effector.
            update_targets (bool): whether to update ik target pos / ori attributes or not

        Returns:
            list: A list of size @num_joints corresponding to the target joint angles.
        """

        # Calculate the rotation
        # This equals: inv base offset * eef * offset accounting for deviation between mujoco eef and pybullet eef
        rotation = self.base_orn_offset_inv @ self.ee_ori_mat @ rotation @ self.rotation_offset[:3, :3]

        # Determine targets based on whether we're using interpolator(s) or not
        if self.interpolator_pos or self.interpolator_ori:
            targets = (self.ee_pos + dpos + self.ik_robot_target_pos_offset, T.mat2quat(rotation))
        else:
            targets = (self.ik_robot_target_pos + dpos, T.mat2quat(rotation))

        # convert from target pose in base frame to target pose in bullet world frame
        world_targets = self.bullet_base_pose_to_world_pose(targets)

        # Update targets if required
        if update_targets:
            # Scale and increment target position
            self.ik_robot_target_pos += dpos

            # Convert the desired rotation into the target orientation quaternion
            self.ik_robot_target_orn = T.mat2quat(rotation)

        # Converge to IK solution
        arm_joint_pos = None
        for bullet_i in range(self.converge_steps):
            arm_joint_pos = self.inverse_kinematics(world_targets[0], world_targets[1])
            self.sync_ik_robot(arm_joint_pos, sync_last=True)

        return arm_joint_pos

    def bullet_base_pose_to_world_pose(self, pose_in_base):
        """
        Convert a pose in the base frame to a pose in the world frame.

        Args:
            pose_in_base (2-tuple): a (pos, orn) tuple.

        Returns:
            2-tuple: a (pos, orn) tuple reflecting robot pose in world coordinates
        """
        pose_in_base = T.pose2mat(pose_in_base)

        base_pos_in_world, base_orn_in_world = p.getBasePositionAndOrientation(
            self.ik_robot, physicsClientId=self.bullet_server_id
        )
        base_pos_in_world, base_orn_in_world = np.array(base_pos_in_world), np.array(base_orn_in_world)

        base_pose_in_world = T.pose2mat((base_pos_in_world, base_orn_in_world))

        pose_in_world = T.pose_in_A_to_pose_in_B(pose_A=pose_in_base, pose_A_in_B=base_pose_in_world)
        return T.mat2pose(pose_in_world)

    def set_goal(self, delta, set_ik=None):
        """
        Sets the internal goal state of this controller based on @delta

        Note that this controller wraps a VelocityController, and so determines the desired velocities
        to achieve the inputted pose, and sets its internal setpoint in terms of joint velocities

        TODO: Add feature so that using @set_ik automatically sets the target values to these absolute values

        Args:
            delta (Iterable): Desired relative position / orientation goal state
            set_ik (Iterable): If set, overrides @delta and sets the desired global position / orientation goal state
        """
        # Update state
        self.update()

        # Get requested delta inputs if we're using interpolators
        (dpos, dquat) = self._clip_ik_input(delta[:3], delta[3:7])

        # Set interpolated goals if necessary
        if self.interpolator_pos is not None:
            # Absolute position goal
            self.interpolator_pos.set_goal(dpos * self.user_sensitivity + self.reference_target_pos)

        if self.interpolator_ori is not None:
            # Relative orientation goal
            self.interpolator_ori.set_goal(dquat)  # goal is the relative change in orientation
            self.ori_ref = np.array(self.ee_ori_mat)  # reference is the current orientation at start
            self.relative_ori = np.zeros(3)  # relative orientation always starts at 0

        # Run ik prepropressing to convert pos, quat ori to desired velocities
        requested_control = self._make_input(delta, self.reference_target_orn)

        # Compute desired velocities to achieve eef pos / ori
        velocities = self.get_control(**requested_control, update_targets=True)

        # Set the goal velocities for the underlying velocity controller
        super().set_goal(velocities)

    def run_controller(self):
        """
        Calculates the torques required to reach the desired setpoint

        Returns:
             np.array: Command torques
        """
        # Update state
        self.update()

        # Update interpolated action if necessary
        desired_pos = None
        rotation = None
        update_velocity_goal = False

        # Update interpolated goals if active
        if self.interpolator_pos is not None:
            # Linear case
            if self.interpolator_pos.order == 1:
                desired_pos = self.interpolator_pos.get_interpolated_goal()
            else:
                # Nonlinear case not currently supported
                pass
            update_velocity_goal = True
        else:
            desired_pos = self.reference_target_pos

        if self.interpolator_ori is not None:
            # Linear case
            if self.interpolator_ori.order == 1:
                # relative orientation based on difference between current ori and ref
                self.relative_ori = orientation_error(self.ee_ori_mat, self.ori_ref)
                ori_error = self.interpolator_ori.get_interpolated_goal()
                rotation = T.quat2mat(ori_error)
            else:
                # Nonlinear case not currently supported
                pass
            update_velocity_goal = True
        else:
            rotation = T.quat2mat(self.reference_target_orn)

        # Only update the velocity goals if we're interpolating
        if update_velocity_goal:
            velocities = self.get_control(dpos=(desired_pos - self.ee_pos), rotation=rotation)
            super().set_goal(velocities)

        # Run controller with given action
        return super().run_controller()

    def update_base_pose(self, base_pos, base_ori):
        # Update pybullet robot base and orientation according to values
        p.resetBasePositionAndOrientation(
            bodyUniqueId=self.ik_robot, posObj=base_pos, ornObj=base_ori, physicsClientId=self.bullet_server_id
        )

        # Re-sync pybullet state
        self.sync_state()

    def update_initial_joints(self, initial_joints):
        # First, update from the superclass method
        super().update_initial_joints(initial_joints)

        # Then, update the rest pose from the initial joints
        self.rest_poses = list(self.initial_joint)

    def reset_goal(self):
        """
        Resets the goal to the current pose of the robot
        """
        self.reference_target_pos = self.ee_pos
        self.reference_target_orn = T.mat2quat(self.ee_ori_mat)

        # Sync pybullet state as well
        self.sync_state()

    def _clip_ik_input(self, dpos, rotation):
        """
        Helper function that clips desired ik input deltas into a valid range.

        Args:
            dpos (np.array): a 3 dimensional array corresponding to the desired
                change in x, y, and z end effector position.
            rotation (np.array): relative rotation in scaled axis angle form (ax, ay, az)
                corresponding to the (relative) desired orientation of the end effector.

        Returns:
            2-tuple:

                - (np.array) clipped dpos
                - (np.array) clipped rotation
        """
        # scale input range to desired magnitude
        if dpos.any():
            dpos, _ = T.clip_translation(dpos, self.ik_pos_limit)

        # Map input to quaternion
        rotation = T.axisangle2quat(rotation)

        # Clip orientation to desired magnitude
        rotation, _ = T.clip_rotation(rotation, self.ik_ori_limit)

        return dpos, rotation

    def _make_input(self, action, old_quat):
        """
        Helper function that returns a dictionary with keys dpos, rotation from a raw input
        array. The first three elements are taken to be displacement in position, and a
        quaternion indicating the change in rotation with respect to @old_quat. Additionally clips @action as well

        Args:
            action (np.array) should have form: [dx, dy, dz, ax, ay, az] (orientation in
                scaled axis-angle form)
            old_quat (np.array) the old target quaternion that will be updated with the relative change in @action
        """
        # Clip action appropriately
        dpos, rotation = self._clip_ik_input(action[:3], action[3:])

        # Update reference targets
        self.reference_target_pos += dpos * self.user_sensitivity
        self.reference_target_orn = T.quat_multiply(old_quat, rotation)

        return {"dpos": dpos * self.user_sensitivity, "rotation": T.quat2mat(rotation)}

    @staticmethod
    def _get_current_error(current, set_point):
        """
        Returns an array of differences between the desired joint positions and current
        joint positions. Useful for PID control.

        Args:
            current (np.array): the current joint positions
            set_point (np.array): the joint positions that are desired as a numpy array

        Returns:
            np.array: the current error in the joint positions
        """
        error = current - set_point
        return error

    @property
    def control_limits(self):
        """
        The limits over this controller's action space, as specified by self.ik_pos_limit and self.ik_ori_limit
        and overriding the superclass method

        Returns:
            2-tuple:

                - (np.array) minimum control values
                - (np.array) maximum control values
        """
        max_limit = np.concatenate([self.ik_pos_limit * np.ones(3), self.ik_ori_limit * np.ones(3)])
        return -max_limit, max_limit

    @property
    def name(self):
        return "IK_POSE"



from PIL import Image
import argparse

import sys
import os

current_file_path = os.path.abspath('')
project_root = os.path.dirname(os.path.dirname(current_file_path))
# 将项目根目录添加到 PYTHONPATH
if project_root not in sys.path:
    sys.path.append(project_root)

from orca_gym.utils.dir_utils import create_tmp_dir, formate_now


def split_channel(input_image_path, save_dir):
    # 打开原始图片
    image = Image.open(input_image_path)

    # 分离出RGB通道
    r, g, b = image.split()

    # 保存分离后的图片
    r.save(save_dir + input_image_path.split('/')[-1].split('.')[0] + '_r.png')
    g.save(save_dir + input_image_path.split('/')[-1].split('.')[0] + '_g.png')
    b.save(save_dir + input_image_path.split('/')[-1].split('.')[0] + '_b.png')
    print("Split channel done!")
    
if __name__ == '__main__':
    parser = argparse.ArgumentParser(description='Helper to Process texture image')
    parser.add_argument('--image', type=str, help='The path to the input image')
    parser.add_argument('--task', type=str, help='The task to run (split_channel / reverse_channel)', default='split_channel')
    args = parser.parse_args()
    
    input_image_path = args.image
    task = args.task
    
    create_tmp_dir('texture_processer_tmp')
    tmp_dir = current_file_path + '/texture_processer_tmp/'
    
    if input_image_path is None:
        print("Please provide the path to the input image")
        exit(1)
    
    if task == 'split_channel':    
        split_channel(input_image_path, tmp_dir)


import sys
import os
import grpc
import numpy as np

class OrcaGymData:
    def __init__(self, model):
        self.model = model
        self.qpos = np.zeros(model.nq)
        self.qvel = np.zeros(model.nv)
        self.qacc = np.zeros(model.nv)
        self.qfrc_bias = np.zeros(model.nv)
        self.time = 0

    def update_qpos_qvel_qacc(self, qpos, qvel, qacc):
        self.qpos = qpos
        self.qvel = qvel
        self.qacc = qacc

    def update_qfrc_bias(self, qfrc_bias):
        self.qfrc_bias = qfrc_bias




import pygame
import sys
import copy
import numpy as np


class XboxJoystick:
    def __init__(self):
        self._joystick_state = None

    def update(self, joystick_state: dict):
        self._joystick_state = joystick_state.copy()

    def get_state(self):
        return self._joystick_state.copy()


    def capture_joystick_pos_ctrl(self) -> dict:
        move_x = self._joystick_state["axes"]["LeftStickX"]
        move_y = -self._joystick_state["axes"]["LeftStickY"]
        move_z = (1 + self._joystick_state["axes"]["RT"]) * 0.5 - (1 + self._joystick_state["axes"]["LT"]) * 0.5
        pos_ctrl = {'x': move_x, 'y': move_y, 'z': move_z}
        return pos_ctrl
    
    def capture_joystick_rot_ctrl(self) -> dict:
        yaw = self._joystick_state["axes"]["RightStickX"]
        pitch = self._joystick_state["axes"]["RightStickY"]
        roll = self._joystick_state["buttons"]["RB"] * 0.5 - self._joystick_state["buttons"]["LB"] * 0.5
        rot_ctrl = {'yaw': yaw, 'pitch': pitch, 'roll': roll}
        return rot_ctrl
    


class XboxJoystickManager:
    def __init__(self):
        # Initialize Pygame
        pygame.init()
        
        # Initialize Joystick
        pygame.joystick.init()

        # Check the operating system
        if sys.platform.startswith('win'):
            self.platform = 'windows'
        elif sys.platform.startswith('linux'):
            self.platform = 'linux'
        else:
            raise Exception("Unsupported platform")

        self._pygame_joysticks = []
        self._xbox_joysticks = []
        joystick_count = pygame.joystick.get_count()
        if joystick_count == 0:
            raise Exception("No joystick detected")
        else:
            for i in range(joystick_count):
                self._pygame_joysticks.append(pygame.joystick.Joystick(i))
                self._pygame_joysticks[-1].init()
                xbox_joystick = XboxJoystick()
                self._xbox_joysticks.append(xbox_joystick)
                print("Joystick detected:", self._pygame_joysticks[-1].get_name())

        # Initialize joystick state
        init_joystick_state = {
            "buttons": {
                "A": 0, "B": 0, "X": 0, "Y": 0,
                "LB": 0, "RB": 0, "Back": 0, "Start": 0,
                "LeftStick": 0, "RightStick": 0
            },
            "axes": {
                "LeftStickX": 0.0, "LeftStickY": 0.0,
                "RightStickX": 0.0, "RightStickY": 0.0,
                "LT": -1.0, "RT": -1.0
            },
            "hats": {
                "DPad": (0, 0)
            }
        }

        self._joystick_states = [copy.deepcopy(init_joystick_state) for i in range(joystick_count)]

        self.BUTTON_MAP = {
            "windows":{
                0: "A", 1: "B", 2: "X", 3: "Y",
                4: "LB", 5: "RB", 6: "Back", 7: "Start",
                8: "LeftStick", 9: "RightStick"},
            "linux":{            
                0: "A", 1: "B", 2: "X", 3: "Y",
                4: "LB", 5: "RB", 6: "Back", 7: "Start",
                8: "LeftStick", 9: "RightStick"}
        }

        self.AXIS_MAP = {
            "windows": {
                0: "LeftStickX", 1: "LeftStickY",
                2: "RightStickX", 3: "RightStickY",
                4: "LT", 5: "RT"},
            "linux": {
                0: "LeftStickX", 1: "LeftStickY",
                3: "RightStickX", 4: "RightStickY",
                2: "LT", 5: "RT"}
        }

        return
    
    def get_joystick_names(self):
        return [joystick.get_name() for joystick in self._pygame_joysticks]

    def get_joystick(self, joystick_name) -> XboxJoystick:
        for i, joystick in enumerate(self._pygame_joysticks):
            if joystick.get_name() == joystick_name:
                return self._xbox_joysticks[i]

        raise Exception("Joystick not found! name: ", joystick_name)
        
    def update(self):
        for event in pygame.event.get():
            # print("event: ", event)
            if event.type == pygame.JOYBUTTONDOWN:
                self._joystick_states[event.joy]["buttons"][self.BUTTON_MAP[self.platform][event.button]] = 1
            if event.type == pygame.JOYBUTTONUP:
                self._joystick_states[event.joy]["buttons"][self.BUTTON_MAP[self.platform][event.button]] = 0
            if event.type == pygame.JOYAXISMOTION:
                if event.axis in self.AXIS_MAP[self.platform]:
                    self._joystick_states[event.joy]["axes"][self.AXIS_MAP[self.platform][event.axis]] = event.value
            if event.type == pygame.JOYHATMOTION:
                self._joystick_states[event.joy]["hats"]["DPad"] = event.value

        for i, joystick in enumerate(self._xbox_joysticks):
            joystick.update(self._joystick_states[i])

    def close(self):
        pygame.quit()
        print("Joysticks closed")



from os import path
from typing import Any, Dict, Optional, Tuple, Union

import numpy as np
from numpy.typing import NDArray

import gymnasium as gym
from gymnasium import error, spaces
from gymnasium.spaces import Space

import asyncio
import sys
from orca_gym import OrcaGymRemote, OrcaGymLocal
from orca_gym.protos.mjc_message_pb2_grpc import GrpcServiceStub 
from orca_gym.utils.rotations import mat2quat, quat2mat

from orca_gym import OrcaGymModel
from orca_gym import OrcaGymData

class RewardType:
    SPARSE = "sparse"
    DENSE = "dense"


class OrcaGymBaseEnv(gym.Env[NDArray[np.float64], NDArray[np.float32]]):
    """Superclass for all OrcaSim environments."""

    def __init__(
        self,
        frame_skip: int,
        orcagym_addr: str,
        agent_names: list[str],
        time_step: float,
        **kwargs
    ):
        """Base abstract class for OrcaSim based environments.

        Args:
            frame_skip: Number of MuJoCo simulation steps per gym `step()`.
            orcagym_addr: The address of the gRPC server.
            agent_names: The names of the agents in the environment.
            time_step: The time step of the simulation.

        Raises:
        """

        # 初始化GRPC通信管道，采用异步通信
        self.orcagym_addr = orcagym_addr
        self.channel = None
        self.stub = None
        self.gym = None
        self._agent_names = agent_names
        self.seed = 0
        self.loop = asyncio.get_event_loop()
        self.initialize_grpc()
        self.pause_simulation()  # 暂停仿真，Gym 采用被动模式，OrcaSim侧不执行周期循环
        self.frame_skip = frame_skip        
        self.set_time_step(time_step)  # 设置仿真时间步长

        self.model, self.data = self.initialize_simulation()
        
        self.reset_simulation() # 重置仿真环境

        self.init_qpos_qvel()

    # methods to override:
    # ----------------------------
    def step(
        self, action: NDArray[np.float32]
    ) -> Tuple[NDArray[np.float64], np.float64, bool, bool, Dict[str, np.float64]]:
        raise NotImplementedError

    def reset_model(self) -> tuple[dict, dict]:
        """
        Reset the robot degrees of freedom (qpos and qvel).
        Implement this in each subclass.
        """
        raise NotImplementedError

    def initialize_simulation(self) -> Tuple[OrcaGymModel, OrcaGymData]:
        """
        Initialize MuJoCo simulation data structures mjModel and mjData.
        """
        raise NotImplementedError

    def _step_orca_sim_simulation(self, ctrl, n_frames) -> None:
        """
        Step over the MuJoCo simulation.
        """
        raise NotImplementedError

    def render(self) -> Union[NDArray[np.float64], None]:
        """
        Render a frame from the MuJoCo simulation as specified by the render_mode.
        """
        raise NotImplementedError

    def generate_action_space(self, bounds : NDArray[np.float64]) -> Space:
        """
        Generate the action space for the environment.
        """
        low, high = 0.0, 0.0
        if len(bounds.T) > 0:
            low, high = bounds.T
        action_space = spaces.Box(low=low, high=high, dtype=np.float32)
        return action_space

    def generate_observation_space(self, obs: Union[Dict[str, Any], np.ndarray]) -> spaces.Space:
        """
        Generate the observation space for the environment.
        """
        if obs is None:
            raise ValueError("obs dictionary is None")

        if isinstance(obs, np.ndarray):
            # If obs is a numpy array, create a box space for it
            low = np.full(obs.shape, -np.finfo(np.float32).max, dtype=np.float32)
            high = np.full(obs.shape, np.finfo(np.float32).max, dtype=np.float32)
            return spaces.Box(low=low, high=high, dtype=np.float32)

        obs_space_dict = {}
        for obs_key, obs_data in obs.items():
            if isinstance(obs_data, np.ndarray):
                # print("obs key: ", obs_key, "dtype : ", obs_data.dtype)
                
                # 创建与 obs_data 形状相同的 low 和 high，并确保它们是 float32
                # 使用 float32 的有限边界值
                finite_max = np.finfo(np.float32).max
                low = np.full(obs_data.shape, -finite_max, dtype=np.float32)
                high = np.full(obs_data.shape, finite_max, dtype=np.float32)
                print(f"low.dtype: {low.dtype}, high.dtype: {high.dtype}")  # 调试信息
                
                obs_space_dict[obs_key] = spaces.Box(
                    low=low,
                    high=high,
                    dtype=obs_data.dtype
                )
            else:
                raise ValueError(f"Unsupported observation type: {type(obs_data)}")
                
        observation_space = spaces.Dict(obs_space_dict)
        return observation_space

    # -----------------------------
    def _get_reset_info(self) -> Dict[str, float]:
        """Function that generates the `info` that is returned during a `reset()`."""
        return {}

    def reset(
        self,
        *,
        seed: Optional[int] = None,
        options: Optional[dict] = None,
    ):
        super().reset(seed=seed)

        if seed is not None:
            self.set_seed_value(seed)

        # mujoco.mj_resetData(self.model, self.data)
        self.reset_simulation()

        obs, info = self.reset_model()

        self.render()
        return obs, info

    def set_seed_value(self, seed=None):
        self.seed_value = seed
        self.np_random = np.random.RandomState(seed)
        return [seed]

    def _name_with_agent0(self, name: str) -> str:
        if len(self._agent_names) > 0:
            return f"{self._agent_names[0]}_{name}"
        else:
            return name
        
    def _name_with_agent(self, agent_id: int, name: str) -> str:
        if len(self._agent_names) > 0:
            return f"{self._agent_names[agent_id]}_{name}"
        else:
            return name

    def body(self, name: str, agent_id = None) -> str:
        if agent_id == None:
            return self._name_with_agent0(name)
        else:
            return self._name_with_agent(agent_id, name)
        
    def joint(self, name: str, agent_id = None) -> str:
        if agent_id == None:
            return self._name_with_agent0(name)
        else:
            return self._name_with_agent(agent_id, name)
        
    
    def actuator(self, name: str, agent_id = None) -> str:
        if agent_id == None:
            return self._name_with_agent0(name)
        else:
            return self._name_with_agent(agent_id, name)
        
    
    def site(self, name: str, agent_id = None) -> str:
        if agent_id == None:
            return self._name_with_agent0(name)
        else:
            return self._name_with_agent(agent_id, name)
        
    
    def mocap(self, name: str, agent_id = None) -> str:
        if agent_id == None:
            return self._name_with_agent0(name)
        else:
            return self._name_with_agent(agent_id, name)

    def sensor(self, name: str, agent_id = None) -> str:
        if agent_id == None:
            return self._name_with_agent0(name)
        else:
            return self._name_with_agent(agent_id, name)
    

    @property
    def dt(self) -> float:
        return self.gym.opt.timestep * self.frame_skip
    
    @property
    def agent_num(self) -> int:
        return len(self._agent_names)

    def do_simulation(self, ctrl, n_frames) -> None:
        """
        Step the simulation n number of frames and applying a control action.
        """
        raise NotImplementedError

    def close(self):
        """Close all processes like rendering contexts"""
        raise NotImplementedError

    def initialize_grpc(self):
        """Initialize the GRPC communication channel."""
        raise NotImplementedError
    
    def pause_simulation(self):
        """Pause the simulation."""
        raise NotImplementedError

    async def _resume_simulation(self):
        """Resume the simulation."""
        raise NotImplementedError
    
    def init_qpos_qvel(self):
        """Init qpos and qvel of the model."""
        raise NotImplementedError
    
    async def _reset_simulation(self):
        """Reset the simulation."""
        raise NotImplementedError
    
    def reset_simulation(self):
        """Reset the simulation."""
        raise NotImplementedError

    def set_time_step(self, time_step):
        """Set the time step of the simulation."""
        raise NotImplementedError
